import sys
sys.path.append("../biped/")
sys.path.append("biped/")
from sympy import *
from biped_dynamics_v1 import Biped_Dynamics
import random
import scipy.integrate
import numpy as np
import collections
import warnings
import time

warnings.filterwarnings("error")




class Gait_Search_Equations( object ):

    """
    holds equations, functions and expressions needed for 
    organizing gait-search
    """
    
    
    def __init__( self, convention, abg, vhc, theta_impact_map ):

        self.abg = abg
        self.vhc = vhc
        self.theta_impact_map = theta_impact_map
	self.convention = convention

    





    # ---------------gait starting conditions constraints---------------------

    # defines constraints for shaping starting conditions
    # for q1 and q2. q3 will then be forced by other constraints 
    # such the biped starts with the given starting q1,q2 and with 
    # both feet on the ground
    
    
    def q_1_init_con_lhs_max( self, theta_0, leg_angle_max ):

        """
        used as inequality constraint
        ensure that q1 - value <=0
        i.e that there is some max abs(q1)
        used to shape start/end cofngis of gait
        """
        
        phi1, phi2, phi3 = self.vhc.get_func_vhc()
        P1, P2, P3 = self.vhc.get_parameterization()

        if self.convention == "absolute":
            
            lhs = -(phi1(theta_0,P1) - leg_angle_max)

        elif self.convention == "relative":
            
            lhs = -(phi1(theta_0,P1) - leg_angle_max)
        
        return lhs


    def q_1_init_con_lhs_min( self, theta_0, leg_angle_min ):

        """
        used as inequality constraint
        ensure that q1 - value >=0
        i.e that there is some min abs(q1)
        used to shape start/end cofngis of gait
        """
        
        phi1, phi2, phi3 = self.vhc.get_func_vhc()
        P1, P2, P3 = self.vhc.get_parameterization()

        if self.convention == "absolute":

            lhs = +(phi1(theta_0,P1) - leg_angle_min)
 
        elif self.convention == "relative":

            lhs = +(phi1(theta_0,P1) - leg_angle_min)

        return lhs

    
    def derive_q_init_grad(self, theta_0):

        phi_1, phi_2, phi_3 = self.vhc.get_sym_vhcs()
        t = symbols('t')
        theta = Function('theta')(t)
        phi_1 = phi_1.subs(theta,theta_0)
        phi_2 = phi_2.subs(theta,theta_0)

        vhc_parameters = self.vhc.get_full_parameter_list()
        
        # create empty matrix
        grad1 = zeros(1,len(vhc_parameters))
        grad2 = zeros(1,len(vhc_parameters))
       
        parameter_count = 0
        for p in vhc_parameters:
            
            grad1[parameter_count] = diff(phi_1,p) 
            grad2[parameter_count] = diff(phi_2,p)
            parameter_count += 1

        # lambdify

        # define lambdifying
        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']
            

        self.q1_min_grad = lambdify( (vhc_parameters,), grad1, modules = mat2array )
        self.q2_min_grad = lambdify( (vhc_parameters,), grad2, modules = mat2array )

        self.q1_max_grad = lambdify( (vhc_parameters,), -grad1, modules = mat2array )
        self.q2_max_grad = lambdify( (vhc_parameters,), -grad2, modules = mat2array )
        

    
    def q_2_init_con_lhs_max( self, theta_0, torso_angle_max ):

        """
        used as inequality constraint
        ensure that q_2 - value <=0
        i.e that there is some max abs(q2)
        used to shape start/end cofngis of gait
        """
        
        phi1, phi2, phi3 = self.vhc.get_func_vhc()
        P1, P2, P3 = self.vhc.get_parameterization()

        if self.convention == "absolute":
            
            lhs = -(phi2(theta_0,P2) - torso_angle_max)

        elif self.convention == "relative":
            
            lhs = -(phi2(theta_0,P2) - torso_angle_max)
        
        return lhs


    
    def q_2_init_con_lhs_min( self, theta_0, torso_angle_min ):

        """
        used as inequality constraint
        ensure that q1 - value >=0
        i.e that there is some min abs(q1)
        used to shape start/end cofngis of gait
        """
        
        phi1, phi2, phi3 = self.vhc.get_func_vhc()
        P1, P2, P3 = self.vhc.get_parameterization()

        if self.convention == "absolute":

            lhs = +(phi2(theta_0,P2) - torso_angle_min)
 
        elif self.convention == "relative":

            lhs = +(phi2(theta_0,P2) - torso_angle_min)

        return lhs
    
    
    
    #---------------------------------------------------------------------------
    
    
    
    
    
    # feet on the ground constraint
    #-----------------------------------------------------------------------------------------------

    def q1_q3_init_symmetry_con_lhs( self, theta_0 ):

        """
        equality constraint ensuring that gait is s
        with both feet on the ground
        """
        phi1, phi2, phi3 = self.vhc.get_func_vhc()
        P1, P2, P3 = self.vhc.get_parameterization()

        if self.convention == "absolute":

            lhs = phi1(theta_0,P1) + phi3(theta_0, P3)

        elif self.convention == "relative":

            lhs = 2*phi1(theta_0,P1) + phi2(theta_0,P2) + phi3(theta_0,P3) - 2*np.pi

        return lhs
    
    def sym_q1_q3_init_symmetry_con_lhs( self, theta_0 ):

        """
        symbolic equality constraint ensuring that gait starts
        with both feet on the ground
        """
        phi_1, phi_2, phi_3 = self.vhc.get_sym_vhcs()

        if self.convention == "absolute":
            
            lhs = phi_1 + phi_3

        elif self.convention == "relative":
            
            lhs = 2*phi_1 + phi_2 + phi_3 - 2*np.pi

        return lhs

    def derive_q1_q3_init_symmetry_con_grad(self, theta_0):
        
        t = symbols('t')
        theta = Function('theta')(t)
        
        lhs = self.sym_q1_q3_init_symmetry_con_lhs(theta_0)
        lhs = lhs.subs(theta,theta_0)

        vhc_parameters = self.vhc.get_full_parameter_list()
        
        # create empty matrix
        grad = zeros(1,len(vhc_parameters))

        parameter_count = 0
        for p in vhc_parameters:
            
            grad[parameter_count] = diff(lhs,p) 
            parameter_count += 1

        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']

        self.q1_q3_init_grad = lambdify( (vhc_parameters,), grad, modules = mat2array )

    
    def eval_q1_q3_init_grad(self,vhc_parameters):
        
        return self.q1_q3_init_grad(vhc_parameters)
   
    
    #-------------------------------------------------------------------------------------------
    
    
    
    
    # gait-symmetry constraint
    #-------------------------------------------------------------------------------------------


    
    def eval_gait_symmetry_con_lhs( self, theta_start, theta_end ):

        
        P1, P2, P3 = self.vhc.get_parameterization()
        P = self.vhc.get_full_parameterization_list()
        
        phi1, phi2, phi3 = self.vhc.get_func_vhc()
        phi1_plus, phi2_plus, phi3_plus = self.theta_impact_map.get_func_phi_plus()
        
        symmetry_constraint_lhs = (phi1(theta_start,P1) - phi1_plus(theta_end,P))**2 + \
                                  (phi2(theta_start,P2) - phi2_plus(theta_end,P))**2 + \
                                  (phi3(theta_start,P3) - phi3_plus(theta_end,P))**2

        return symmetry_constraint_lhs



    def derive_gait_symmetry_con_grad( self, theta_start, theta_end ):
	
        t = symbols('t')
        theta = Function('theta')(t)

        phi_1_plus, phi_2_plus, phi_3_plus = self.theta_impact_map.get_sym_phi_plus()

        phi_1, phi_2, phi_3 = self.vhc.get_sym_vhcs()

        phi_1_start = phi_1.subs(theta, theta_start)
        phi_2_start = phi_2.subs(theta, theta_start)
        phi_3_start = phi_3.subs(theta, theta_start)

        phi_1_end_plus = phi_1_plus.subs(theta,theta_end)
        phi_2_end_plus = phi_2_plus.subs(theta,theta_end)
        phi_3_end_plus = phi_3_plus.subs(theta,theta_end)

        symmetry_constraint_lhs = (phi_1_start - phi_1_end_plus)**2 + \
                                  (phi_2_start - phi_2_end_plus)**2 + \
                                  (phi_3_start - phi_3_end_plus)**2

        
        vhc_parameters = self.vhc.get_full_parameter_list()
        
        # create empty matrix
        grad = zeros(1,len(vhc_parameters))

        parameter_count = 0
        for p in vhc_parameters:
            
            grad[parameter_count] = diff(symmetry_constraint_lhs,p) 
            parameter_count += 1

        self.sym_gait_symmetry_grad = grad

        # lambdify

        # define lambdifying
        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']
        
        self.gait_symmetry_constraint_grad = lambdify( (vhc_parameters,), grad, modules = mat2array )


    def eval_gait_symmetry_con_grad( self, vhc_parameters ):
        
        return self.gait_symmetry_constraint_grad(vhc_parameters)

    #----------------------------------------------------------------------------------------------------------


        

    
    
    # theta-impact
    #----------------------------------------------------------------------------------------------------------
    
    
    
    def eval_theta_impact_obj_func( self ): 

        k1, k2, k3 = self.theta_impact_map.get_factors()

        return (k1 - k2)**2 + (k1 - k3)**2


    def sym_impact_obj_func( self ):
        
        k_1, k_2, k_3 = self.theta_impact_map.get_sym_factors()

        return (k_1 - k_2)**2 + (k_1 - k_3)**2




    def derive_theta_impact_obj_grad( self ):
        
        obj_func = self.sym_impact_obj_func()

        vhc_parameters = self.vhc.get_full_parameter_list()

        grad = zeros(1, len(vhc_parameters))

        parameter_count = 0
        for p in vhc_parameters:
            
            grad[parameter_count] = diff(obj_func,p) 
            parameter_count += 1


        # lambdify

        # define lambdifying
        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']

        self.theta_impact_obj_grad = lambdify( (vhc_parameters,), grad, modules = mat2array )

    
    def eval_theta_impact_obj_grad( self ):
        
        P = self.vhc.get_full_parameterization_list()
        
        grad = self.theta_impact_obj_grad(P)

        return grad
    
    #----------------------------------------------------------------------------------------------------




    
    
    
    #torque equations
    # ---------------------------------------------------------------------------------------------------

    
    def eval_torque_con_lhs( self, theta_start, theta_end, theta_d_0_sq, theta_d_t_sq ):

        """
        returns lhs of torque equality constraint equation
        constraint_func = 0
        """

        theta_start_dot = np.sign(theta_d_0_sq)*np.sqrt(np.abs(theta_d_0_sq))
        theta_end_dot = np.sign(theta_d_t_sq)*np.sqrt(np.abs(theta_d_t_sq))
        

        P = self.vhc.get_full_parameterization_list()

        tau_2_start = self.abg.tau2(theta_start, theta_start_dot, P)
        tau_3_start = self.abg.tau3(theta_start, theta_start_dot, P)

        tau_2_end = self.abg.tau2(theta_end, theta_end_dot, P)
        tau_3_end = self.abg.tau3(theta_end, theta_end_dot, P)
        
        try:
            eq_constr_lhs = (tau_2_start - tau_3_end)**2 + (tau_3_start - tau_2_end)**2
        
        except Warning as eval_torque_warn:
            
            eq_constr_lhs = 10000
            #print "something went wrong in eval_torque_con_lhs"
            #print eval_torque_warn, "\n"
        
        return eq_constr_lhs
        


    def derive_torque_con_grad( self, theta_start, theta_end, theta_start_dot, theta_end_dot ):

        """
        Derives the analytical gradient for the torque 
        continuity constraint
        """

        t = symbols('t')
        theta = Function('theta')(t)
        theta_dot = diff(theta,t)
        
        tau_2 = self.abg.sym_tau_2()
        tau_3 = self.abg.sym_tau_3()

        tau_2_start = tau_2.subs([ (theta_dot, theta_start_dot), (theta, theta_start) ])
        tau_3_start = tau_3.subs([ (theta_dot, theta_start_dot), (theta, theta_start) ])

        tau_2_end = tau_2.subs([ (theta_dot, theta_end_dot), (theta, theta_end) ])
        tau_3_end = tau_3.subs([ (theta_dot, theta_end_dot), (theta, theta_end) ])

        # the equality constraint rhs

        eq_constr_lhs = (tau_2_start - tau_3_end)**2 + (tau_3_start - tau_2_end)**2 
        
        vhc_parameters = self.vhc.get_full_parameter_list()
        
        # create empty matrix
        grad = zeros(1,len(vhc_parameters))

        parameter_count = 0
        for p in vhc_parameters:

            grad[parameter_count] = diff(eq_constr_lhs,p) 
            parameter_count += 1
	

        self.sym_toque_constr_gradient = grad

        # lambdify
        # define lambdifying
        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']
        
        self.torque_constraint_grad = lambdify( (vhc_parameters,), grad, modules = mat2array )

    
    
    def eval_torque_con_grad( self, vhc_parameters ):

        return self.torque_constraint_grad( vhc_parameters )


    def eval_max_torque_con( self, theta, max_value ):

        return max_value - self.max_torque(theta) 



    def max_torque( self, theta ):

        tau2 = self.abg.tau2
        tau3 = self.abg.tau3

        P = self.vhc.get_full_parameterization_list()

        rho1, rho2 = self.fast_rho1_rho2_arr( theta )

        theta_d_sq = self.fast_theta_d_sq( rho1, rho2 )

        t2 = []
        t3 = []

        if np.amin(theta_d_sq) > 0:

            theta_d = np.sqrt(theta_d_sq)

            for i in range( 0, len(theta) ):

                t2.append( abs(tau2( theta[i], theta_d[i], P )) )
                t3.append( abs(tau3( theta[i], theta_d[i], P )) )

            return max( max(t2), max(t3) )

        else:

            return 100


    

    #------------------------------------------------------------------------------------------------------------

    
    # alpha < 0
    def max_alpha(self, theta):

        P = self.vhc.get_full_parameterization_list()
        
        alpha_fun = self.abg.alpha0
        
        alpha = []
        
        for th in theta:

            alpha.append(alpha_fun(th,P))

        return max(alpha)


    def eval_max_alpha_con(self, theta, offset):

        return - (self.max_alpha(theta) + offset)


    
    
    
    # motion integral and theta_dot calculations 
    #----------------------------------------------------------------


    # fast integral calculation implementations 
    #---------------------------------------------------------------
    
    def fast_rho1_rho2_arr( self, theta_vec ):

        """
        returns a numpy array of the cumulative rho1 and rho2
        """
        
        alpha, beta, gamma = self.abg.get_func_abg_0()
        P = self.vhc.get_full_parameterization_list()
        
        def integrand(tau):
            return - (2.0*beta(tau,P))/(alpha(tau,P)) 
        
        def g_div_a(s):
            return (2.0*gamma(s,P))/alpha(s,P)
        
        try: 
            integrand1 = np.array( map(integrand,theta_vec) )

            psi = scipy.integrate.cumtrapz( integrand1, theta_vec, initial = 0 )  

            rho1 = np.exp(psi)

            integrand3 = np.exp(np.negative(psi))*np.array( map(g_div_a,theta_vec) )

            rho2 = scipy.integrate.cumtrapz( integrand3, theta_vec, initial = 0 )*rho1


        except Warning as rho_warn:

            rho1 = np.array([False])
            rho2 = np.array([False])
        
        return rho1, rho2
   
    
     
    def fast_rho1_rho2_t( self, theta_vec ):

        """
        returns end values (scalar) of the cumulative rho1 and rho2
        """

        
        alpha, beta, gamma = self.abg.get_func_abg_0()
        P = self.vhc.get_full_parameterization_list()
        
        def integrand(tau):
            return - (2.0*beta(tau,P))/(alpha(tau,P)) 
        
        def g_div_a(s):
            return (2.0*gamma(s,P))/alpha(s,P)
        

        try:
            
            integrand1 = np.array( map(integrand,theta_vec) )

            psi = scipy.integrate.cumtrapz( integrand1, theta_vec, initial = 0 )  

            rho1 = np.exp(psi)

            psi = psi[-1] - psi

            integrand2 = np.exp(psi)*np.array( map(g_div_a,theta_vec) )

            rho2 = scipy.integrate.cumtrapz( integrand2, theta_vec, initial = 0 )

            return rho1[-1], rho2[-1]
   
    
        except Warning as integration_warn:

            #print "something went wrong in fast_mi :"
            #print integration_warn, "\n"
            
            
            rho1 = np.array([False])
            rho2 = np.array([False])

            return rho1, rho2
  
    
    
    def fast_theta_d_sq( self, rho1_arr, rho2_arr ):

        """
        calculates the theta-velocity squard from theta_0 theta_t 
        using the motion integral 
        """

        theta_d_0_sq = self.fast_theta_d_0_sq(rho1_arr, rho2_arr)
                
        try:
            
            theta_d_t_sq = rho1_arr*theta_d_0_sq - rho2_arr
            return theta_d_t_sq

        except Warning as theta_d_warn:

            print theta_d_warn
            return np.array([-1000])

    
    
    def min_theta_d_sq( self, rho1_arr, rho2_arr ):

        """
        returns the minumum value of theta velocity squared
        along theta from theta_0 to theta_t
        """
        
        theta_d_sq = self.fast_theta_d_sq( rho1_arr, rho2_arr )

        return np.amin(theta_d_sq)
        
    
    
    
    def fast_theta_d_0_sq( self, rho1_arr, rho2_arr ):

        """
        calculates initial velocity of theta 
        from the impact-map and Integral equation
        k2 is chosen arbitraily as k1=k2=k3 when
        search procedure converges on a walking gait
        """

        k1, k2, k3 = self.theta_impact_map.get_factors()
        k = k2
        
        rho1, rho2 = rho1_arr[-1], rho2_arr[-1]
       
        
        if rho1 != -1 and rho2 != -1:
            
            try:
                theta_d_0_sq = (rho2*(k**2))/(rho1*(k**2) - 1)
                
                return theta_d_0_sq 
            
            except Warning as theta_d_warn:

                #print "something went wrong in fast_theta_d_0_sq: "
                #print theta_d_warn, "\n"
                
                return -100, -100
        
        else:
            return -100, -100

     
    def fast_theta_d_t_sq( self, rho1_arr, rho2_arr ):
        
        """
        calculates theta_d_t_sq, the end velocity from the impact-map and
        Integral equation
        """

        k1, k2, k3 = self.theta_impact_map.get_factors()
        k = k2
        
        rho1, rho2 = rho1_arr[-1], rho2_arr[-1]       
        
        if rho1 != -1 and rho2 != -1:
            
            try:
                
                theta_d_t_sq = (rho2)/(rho1*(k**2) - 1)
                
                return theta_d_t_sq 
            
            except Warning as theta_d_warn:

                #print "something went wrong in fast_theta_d_0_sq: "
                #print theta_d_warn, "\n"
                
                return -100, -100
        
        else:
            return -100, -100



    def fast_theta_d_0_t_sq( self, rho1_arr, rho2_arr ):

        """
        claculates and return theta_d_0_sq and theta_d_t_sq from 
        impact-map and motion integral equation
        """

        k1, k2, k3 = self.theta_impact_map.get_factors()
        k = k2
        
        rho1, rho2 = rho1_arr[-1], rho2_arr[-1]
        
        if rho1 != -1 and rho2 != -1:
            
            try:
                
                theta_d_0_sq = (rho2*(k**2))/(rho1*(k**2) - 1)
                theta_d_t_sq = (rho2)/(rho1*(k**2) - 1)

                return theta_d_0_sq, theta_d_t_sq 
            
            except Warning as theta_d_warn:

                #print "something went wrong in fast_theta_d_0_sq: "
                #print theta_d_warn, "\n"
                
                return -100, -100
        
        else:
            return -100, -100

    
    
    def motion_integral_eq(self, rho1_arr, rho2_arr, theta_d_0_sq, theta_d_t_sq ):

        """
        calculates the motion Integral of entire theta renging
        from theta_0 to theta_t
        """

        rho1, rho2 = rho1_arr, rho2_arr

        try:
                
            I = theta_d_t_sq - rho1*theta_d_0_sq + rho2
            return I 
            
        except Warning as theta_d_warn:

            #print "something went wrong in fast_theta_d_0_sq: "
            #print theta_d_warn, "\n"
                
            return 10000
        
    


    


# slow calculations of Integral and theta_dot
#-----------------------------------------------------------------------------------------------



    def slow_motion_integral( self, theta_0, theta_t, P ):

        """
        calculates motion-integrals in a slow way 
        using scipy.integrate.quad
        """


        warnings.filterwarnings("error")
            
        abg = self.abg 
        alpha, beta, gamma = abg.get_func_abg_0()
        

        def integrand1(tau):
            return - (2*beta(tau,P))/(alpha(tau,P))
            
        def integrand2(tau):
            return (2*beta(tau,P))/(alpha(tau,P))

        def integrand3(s):
            return np.exp( scipy.integrate.quad(integrand2,theta_t,s, full_output = 1 )[0] )\
                          *( (2.0*gamma(s,P))/alpha(s,P) )
        
        try:
            
            rho1 = np.exp( scipy.integrate.quad( integrand1, theta_0, theta_t )[0] )
            rho2 = scipy.integrate.quad( integrand3, theta_0, theta_t )[0] 

            return rho1, rho2
        
        except Warning as integration_warn:
            
            print "something went wrong in Motion_Intgral:  \n" 
            print integration_warn
            
            rho1 = -1
            rho2 = -1

            return rho1, rho2



    def slow_theta_d_0_t_sq(self, theta_0, theta_t, P):

        """
        returns starting and ending velocitites of theta
        """
        
        k1, k2, k3 = self.theta_impact_map.get_factors()
        k = k2
        rho1, rho2 = self.slow_motion_integral(theta_0, theta_t, P )

        
        
        if rho1 != -1 and rho2 != -1:
            
            try:
                theta_d_0_sq = (rho2*(k**2))/(rho1*(k**2) - 1)
                theta_d_t_sq = (rho2)/(rho1*(k**2) - 1)
                
                return theta_d_0_sq, theta_d_t_sq 
            
            except Warning as theta_d_warn:

                #print "something went wrong in theta_d_0_sq: "
                #print theta_d_warn, "\n"
                
                return -100, -100
        
        else:
            return -100, -100


# ----------------------------------------------------------------------------------------------------------------


