import sys
sys.path.append("../q_space_dynamics/")
sys.path.append("../theta_space_dynamics/")

from biped_data import Biped_Data
from sympy import *
import numpy as np
import lxml.etree as ET


class Q_Space_Kinematics( Biped_Data ):

    
    def __init__( self, convention, model_xml ):


        
        super( Q_Space_Kinematics, self ).__init__()
        
        self.load_Model_XML(model_xml)
         
        if convention == "absolute":
            self.absolute_angles()
        
        elif convention == "relative":
            self.relative_angles()
        
        else:
            print "WARNING: No valid angle convention given to Biped_Dynamics"
            self.absolute_angles()



    def absolute_angles( self ):

        """
        Derivation of kinematics in absolute angles        
        """
        
        #--------- Symbolic Definitions -----------------------

        #define names of physical parameters as symbols
        l_1 = symbols('l_1')
        l_2 = symbols('l_2')
        l_3 = symbols('l_3')

        L_1 = symbols('L_1')
        L_2 = symbols('L_2')
        L_3 = symbols('L_3')

        m_1 = symbols('m_1')
        m_2 = symbols('m_2')
        m_3 = symbols('m_3')

        J_1 = symbols('J_1')
        J_2 = symbols('J_2')
        J_3 = symbols('J_3')

        g = symbols('g')


        #define generalized coordinates as functions of t
        #"Function" is a method from sympy module defining q_i to be functions of t
        t=symbols('t')
        q_1=Function('q_1')(t)
        q_2=Function('q_2')(t)
        q_3=Function('q_3')(t)

        #define dot(q_i) and ddot(q_i) to be symbols Dq_i and DDq_i
        Dq_1 = diff(q_1,t)
        Dq_2 = diff(q_2,t)
        Dq_3 = diff(q_3,t)
        DDq_1 = diff(q_1,t)
        DDq_2 = diff(q_2,t)
        DDq_3 = diff(q_3,t)

        # symbolic state vectors
        q = Matrix([[q_1],[q_2],[q_3]]) 
        Dq = Matrix([[Dq_1],[Dq_2],[Dq_3]])
        #-----------------------------------------------------


        #--------- Kinematic Relations -----------------------

        # Positions
        
        p_1=Matrix([
            [-l_1*sin(q_1)],
            [l_1*cos(q_1)]
            ])
        
        p_2=Matrix([
            [-L_1*sin(q_1)-l_2*sin(q_2)],
            [L_1*cos(q_1)+l_2*cos(q_2)]
            ])
        
        p_3=Matrix([
            [-L_1*sin(q_1)+l_3*sin(q_3)],
            [L_1*cos(q_1)-l_3*cos(q_3)]
            ])
        
        p_3f=Matrix([
            [-L_1*sin(q_1)+L_3*sin(q_3)],
            [L_1*cos(q_1)-L_3*cos(q_3)]
            ])


        #Linear Velocities
        v_1=p_1.diff(t)
        v_2=p_2.diff(t)
        v_3=p_3.diff(t)
        v_3f=p_3f.diff(t)


        # setting symbolic kin as instance var
        self.p_1, self.p_2, self.p_3, self.p_3f = p_1, p_2, p_3, p_3f
        self.v_1, self.v_2, self.v_3, self.v_3f = v_1, v_2, v_3, v_3f
        
        #------------------------------------------------------

    
    
    def relative_angles( self ):
        
        """
        Derivation of dynamics in relative angles        
        """
   	
       
        #--------- Symbolic Definitions -----------------------

        #define names of physical parameters as symbols
        l_1 = symbols('l_1')
        l_2 = symbols('l_2')
        l_3 = symbols('l_3')

        L_1 = symbols('L_1')
        L_2 = symbols('L_2')
        L_3 = symbols('L_3')

        m_1 = symbols('m_1')
        m_2 = symbols('m_2')
        m_3 = symbols('m_3')

        J_1 = symbols('J_1')
        J_2 = symbols('J_2')
        J_3 = symbols('J_3')

        g = symbols('g')


        #define generalized coordinates as functions of t
        #"Function" is a method from sympy module defining q_i to be functions of t
        t=symbols('t')
        q_1=Function('q_1')(t)
        q_2=Function('q_2')(t)
        q_3=Function('q_3')(t)

        #define dot(q_i) and ddot(q_i) to be symbols Dq_i and DDq_i
        q_1_d = diff(q_1,t)
        q_2_d = diff(q_2,t)
        q_3_d = diff(q_3,t)

        # symbolic state vectors
        Q = Matrix([[q_1],[q_2],[q_3]]) 
        Q_d = Matrix([[q_1_d],[q_2_d],[q_3_d]])
        
        self.Q = Q
        self.Q_d = Q_d
        #-----------------------------------------------------


        #--------- Kinematic Relations -----------------------

        #Position 
        p_1=Matrix([
            [l_1*cos(q_1)],
            [l_1*sin(q_1)]
            ])
        p_2=Matrix([
            [L_1*cos(q_1)+l_2*cos(q_1+q_2)],
            [L_1*sin(q_1)+l_2*sin(q_1+q_2)]
            ])
        p_3=Matrix([
            [L_1*cos(q_1)+l_3*cos(q_1+q_2+q_3)],
            [L_1*sin(q_1)+l_3*sin(q_1+q_2+q_3)]
            ])  
        p_3f=Matrix([
            [L_1*cos(q_1)+L_3*cos(q_1+q_2+q_3)],
            [L_1*sin(q_1)+L_3*sin(q_1+q_2+q_3)]
            ])

        #Linear Velocities
        v_1=p_1.diff(t)
        v_2=p_2.diff(t)
        v_3=p_3.diff(t)
        v_3f=p_3f.diff(t)
        
        # setting symbolic kin as instance var
        self.p_1, self.p_2, self.p_3, self.p_3f = p_1, p_2, p_3, p_3f
        self.v_1, self.v_2, self.v_3, self.v_3f = v_1, v_2, v_3, v_3f

        #------------------------------------------------------

        
    def get_sym_position( self ):

        return self.p_1, self.p_2, self.p_3,  self.p_3f

    
    def get_sym_velocity( self ):

        return self.v_1, self.v_2, self.v_3,  self.v_3f


