import lxml.etree as ET

# ----------------------------------------------------------------

class Biped_Data(object) :
    
    """
    A biped data holder structure for passing the data 
    into the main Biped class. 
    
    It is used for data handling and parsing to/from XML files 
    """
    
    def __init__( self ) :
        
        # physical constants
        self.g  =  9.81

        # a default set of parameters

        m1 = 1.0
        m2 = 2.0
        m3 = 1.0

        self.m1 = m1
        self.m2 = m2
        self.m3 = m3

        self.m2_data = m2

        L1 = 1.0
        L2 = 2.0
        L3 = 1.0

        self.L1 = L1
        self.L2 = L2
        self.L3 = L3

        l1 = L1*0.5
        l2 = L2*0.5
        l3 = L3*0.5

        self.l1 = l1
        self.l2 = l2
        self.l3 = l3

        self.l2_data = l2

        J1 = m1 * L1**2 / 12
        J2 = m2 * L2**2 / 12
        J3 = m3 * L3**2 / 12

        self.J1 = J1
        self.J2 = J2
        self.J3 = J3

        self.J2_data = J2
        
        self.Mh = 0.0         
        
        self.derived_mass_parameters( )

    def set_data( self, m1, m2, m3, L1, L2, L3, Mh=0.0 ) :
        
        """ 
        Method sets up the physical data of the biped model 
        assuming uniform mass distribution in the links 
        as a result CoMs of the links are in the middle l_i = L_i*0.5
        and J_i = ( m_i * L_i**2 )/12 
        """
        
        self.m2_data = m2        
        
        l1 = L1 * 0.5        
        l2 = L2 * 0.5        
        l3 = L3 * 0.5        
        
        self.l2_data = l2

        # this inertia formula assumes uniform mass distribution 
        # as a result li = Li*0.5
        J1 = ( m1 * L1**2 )/12 
        J2 = ( m2 * L2**2 )/12
        J3 = ( m3 * L3**2 )/12

        self.J2_data = J2 
        
        lH2 = l2 * m2 /( m2+Mh )

        self.Mh = Mh

        self.m1 = m1
        self.m2 = m2 + Mh
        self.m3 = m3

        self.L1 = L1
        self.L2 = L2
        self.L3 = L3

        self.l1 = l1   
        self.l2 = lH2 
        self.l3 = l3  

        self.J1 = J1 
        self.J2 = J2 + Mh*lH2**2 + m2*(l2-lH2)**2 
        self.J3 = J3
        
        self.derived_mass_parameters( )

    def derived_mass_parameters( self ) :         
        
        # main parameters
        g, m1, m2, m3, L1, L2, L3, l1, l2, l3, J1, J2, J3 = self.get_data()          

        # calculating the derived parameters to avoid repeated calculations  
        self.JLM1 = L1**2*m2 + L1**2*m3 + l1**2*m1 + l2**2*m2 + l3**2*m3 + J1 + J2 + J3
        self.JLM2 = l2**2*m2 + l3**2*m3 + J2 + J3
        self.JLM3 = l3**2*m3 + J3
    
    def get_data( self ) :

        """getting the main physical parameters as a tuple"""
        
        return self.g, self.m1, self.m2, self.m3, self.L1, self.L2, self.L3, self.l1, self.l2, self.l3, self.J1, self.J2, self.J3

    def get_derived_data( self ) : 

        """getting the derived physical parameters as a tuple""" 

        return self.JLM1, self.JLM2, self.JLM3
        
    def get_Element( self ) :

        """
        generating the root element of the XML data file storing the physical 
        parameters of the biped 
        ... very flexible for adding extra data structures for VHC trajectory ++ 
        by redefining this method in the children classes and calling the 
        parent class method in the child class equivalent 
        """
        
        et_root = ET.Element( "biped_model" ) 
   
        phys_params = ET.SubElement( et_root, "physical_parameters" )     
        
        if hasattr( self, "description" ) :
            descrpt = ET.Element("description")
            descrpt.text = self.description
            phys_params.append( descrpt )
        
        g_data = ET.SubElement( phys_params, "gravitational_constant" )        
        g_data.set( "model_parameter", "g" )        
        g_data.text = str( self.g )        
        
        stance_leg = ET.SubElement( phys_params, "stance_leg_parameters" )  
        
        m1 = ET.SubElement( stance_leg, "mass" )        
        m1.set( "model_parameter", "m1" )        
        m1.text = str( self.m1 )        
        
        L1 = ET.SubElement( stance_leg, "length" )        
        L1.set( "model_parameter", "L1" )        
        L1.text = str( self.L1 )        

        l1 = ET.SubElement( stance_leg, "CoM_position" )        
        l1.set( "model_parameter", "l1" )        
        l1.text = str( self.l1 )        

        J1 = ET.SubElement( stance_leg, "inertia" )        
        J1.set( "model_parameter", "J1" )        
        J1.text = str( self.J1 )        

        body = ET.SubElement( phys_params, "body_parameters" )     

        m2 = ET.SubElement( body, "mass" )        
        m2.set( "model_parameter", "m2" )        
        m2.text = str( self.m2_data )        
        
        L2 = ET.SubElement( body, "length" )        
        L2.set( "model_parameter", "L2" )        
        L2.text = str( self.L2 )        

        l2 = ET.SubElement( body, "CoM_position" )        
        l2.set( "model_parameter", "l2" )        
        l2.text = str( self.l2_data )        

        J2 = ET.SubElement( body, "inertia" )        
        J2.set( "model_parameter", "J2" )        
        J2.text = str( self.J2_data )        

        swing_leg  = ET.SubElement( phys_params, "swing_leg_parameters" )     

        m3 = ET.SubElement( swing_leg, "mass" )        
        m3.set( "model_parameter", "m3" )        
        m3.text = str( self.m3 )        
        
        L3 = ET.SubElement( swing_leg, "length" )        
        L3.set( "model_parameter", "L3" )        
        L3.text = str( self.L3 )        

        l3 = ET.SubElement( swing_leg, "CoM_position" )        
        l3.set( "model_parameter", "l3" )        
        l3.text = str( self.l3 )        

        J3 = ET.SubElement( swing_leg, "inertia" )        
        J3.set( "model_parameter", "J3" )        
        J3.text = str( self.J3 )        

        hips_mass  = ET.SubElement( phys_params, "hips_mass_parameter" )     

        Mh = ET.SubElement( hips_mass, "mass" )        
        Mh.set( "model_parameter", "Mh" )        
        Mh.text = str( self.Mh )        

        return et_root
    
    def save_Model_XML( self, xml_file_name ) : 
        
        """
        method: save_Model_XML( xml_file_name )
        saving the model data as an XML file
        """
        
        tree = ET.ElementTree( self.get_Element() )
        
        tree.write( xml_file_name, pretty_print=True )

    def set_data_from_RootElement( self, root_element ) : 
        
        """
        data parsing from the root element of the XML file 
        containing the pysical parameters 
        ... very flexible for adding extra data to the data file format
        by adding subsequent parser functionality in the children classes 
        """
        
        phys_params = root_element.find( "physical_parameters" )
        
        # description text of the parameters set can be added 
        descript = phys_params.find( "description" )
        if descript is not None :
            self.description = descript.text        
        
        self.g = float( phys_params.find( "gravitational_constant" ).text ) 

        stance_leg = phys_params.find( "stance_leg_parameters" )
        self.m1 = float( stance_leg.find( "mass" ).text )
        self.L1 = float( stance_leg.find( "length" ).text )
        self.l1 = float( stance_leg.find( "CoM_position" ).text )
        self.J1 = float( stance_leg.find( "inertia" ).text )
        
        body = phys_params.find( "body_parameters" )
        m2 = float( body.find( "mass" ).text )
        L2 = float( body.find( "length" ).text )
        l2 = float( body.find( "CoM_position" ).text )
        J2 = float( body.find( "inertia" ).text )

        # stored separately in case Mh is not 0.0
        self.m2_data = m2
        self.L2 = L2
        self.l2_data = l2
        self.J2_data = J2

        swing_leg = phys_params.find( "swing_leg_parameters" )
        self.m3 = float( swing_leg.find( "mass" ).text )
        self.L3 = float( swing_leg.find( "length" ).text )
        self.l3 = float( swing_leg.find( "CoM_position" ).text )
        self.J3 = float( swing_leg.find( "inertia" ).text )

        hips_mass = phys_params.find( "hips_mass_parameter" )
        Mh = float( hips_mass.find( "mass" ).text )
        
        self.Mh = Mh
        
        # even if Mh == 0.0 
        lH2 = l2 * m2 /( m2+Mh )

        self.m2 = m2 + Mh
        self.l2 = lH2
        self.J2 = J2 + Mh*lH2**2 + m2*(l2-lH2)**2

        self.derived_mass_parameters( )

    def load_Model_XML( self, xml_file_name ) : 
        
        """
        method: load_Model_XML( xml_file_name )
        loading the model data from an XML file
        """

        # getting the root element of the XML document 
        model = ET.parse( xml_file_name ).getroot()
        
        # passing it to the parser
        self.set_data_from_RootElement( model )
