from biped_data import Biped_Data 
from sympy import *
import numpy as np
import lxml.etree as ET

class Biped_Dynamics(Biped_Data):

    def __init__( self, convention, model_xml ):


        super( Biped_Dynamics, self ).__init__()    

        if convention == "absolute":
            self.load_Model_XML(model_xml)
            self.absolute_angles()
        
        elif convention == "relative":
            self.load_Model_XML(model_xml)
            self.relative_angles()
        
        else:
            print "WARNING: No valid angle convention given to Biped_Dynamics"
            self.load_Model_XML(model_xml)
            self.absolute_angles()
        



    def absolute_angles( self ):

        

        """
        Derivation of dynamics in absolute angles        
        """
    	
        
        #--------- Symbolic Definitions -----------------------

        #define names of physical parameters as symbols
        l_1 = symbols('l_1')
        l_2 = symbols('l_2')
        l_3 = symbols('l_3')

        L_1 = symbols('L_1')
        L_2 = symbols('L_2')
        L_3 = symbols('L_3')

        m_1 = symbols('m_1')
        m_2 = symbols('m_2')
        m_3 = symbols('m_3')

        J_1 = symbols('J_1')
        J_2 = symbols('J_2')
        J_3 = symbols('J_3')

        g = symbols('g')


        #define generalized coordinates as functions of t
        #"Function" is a method from sympy module defining q_i to be functions of t
        t=symbols('t')
        q_1=Function('q_1')(t)
        q_2=Function('q_2')(t)
        q_3=Function('q_3')(t)

        #define dot(q_i) and ddot(q_i) to be symbols Dq_i and DDq_i
        Dq_1 = diff(q_1,t)
        Dq_2 = diff(q_2,t)
        Dq_3 = diff(q_3,t)
        DDq_1 = diff(q_1,t)
        DDq_2 = diff(q_2,t)
        DDq_3 = diff(q_3,t)

        # symbolic state vectors
        q = Matrix([[q_1],[q_2],[q_3]]) 
        Dq = Matrix([[Dq_1],[Dq_2],[Dq_3]])
        #-----------------------------------------------------


        #--------- Kinematic Relations -----------------------

        #Position 
        p_1=Matrix([
            [-l_1*sin(q_1)],
            [l_1*cos(q_1)]
            ])
        p_2=Matrix([
            [-L_1*sin(q_1)-l_2*sin(q_2)],
            [L_1*cos(q_1)+l_2*cos(q_2)]
            ])
        p_3=Matrix([
            [-L_1*sin(q_1)+l_3*sin(q_3)],
            [L_1*cos(q_1)-l_3*cos(q_3)]
            ])

        #Linear Velocities
        v_1=p_1.diff(t)
        v_2=p_2.diff(t)
        v_3=p_3.diff(t)

        #------------------------------------------------------


        #--------- Kinetic and Potential Energy ---------------


        # Stance leg
        K_1 = (m_1/2)*(v_1[0]**2+v_1[1]**2) + (J_1/2)*(Dq_1**2)
        P_1 = m_1*g*p_1[1]

        # Torso
        K_2 = (m_2/2)*(v_2[0]**2+v_2[1]**2) + (J_2/2)*(Dq_2**2)
        P_2 = m_2*g*p_2[1]

        # Swing leg
        K_3 = (m_3/2)*(v_3[0]**2 + v_3[1]**2) + (J_3/2)*(Dq_3**2)
        P_3 = m_3*g*p_3[1]

        # Collecting total kinetic and potential energy
        K = K_1 + K_2 + K_3
        P = P_1 + P_2 + P_3
        
        
        #-----------------------------------------------------

        #--------- Euler-Lagrange equations of motion --------
        
        
        L = K - P

        mellom_1 = diff(K,Dq_1)
        mellom_2 = diff(K,Dq_2)
        mellom_3 = diff(K,Dq_3)
        
        LHS1_alt = diff(mellom_1,t) - diff(K,q_1) + diff(P,q_1)
        LHS2_alt = diff(mellom_2,t) - diff(K,q_2) + diff(P,q_2)
        LHS3_alt = diff(mellom_3,t) - diff(K,q_3) + diff(P,q_3)
     
        # create LHS-expressions
        LHS1 = diff(diff(K,Dq_1),t) - diff(K,q_1) + diff(P,q_1)
        LHS2 = diff(diff(K,Dq_2),t) - diff(K,q_2) + diff(P,q_2)
        LHS3 = diff(diff(K,Dq_3),t) - diff(K,q_3) + diff(P,q_3)

        self.LHS1 = LHS1_alt
        self.LHS2 = LHS2_alt
        self.LHS3 = LHS3_alt

        # derive matrices M(q),C(q,Dq),G(q)

        # M(q)
        M = zeros(3,3)
        for i in range(0,3):
            for j in range(0,3):
                M[i,j] = diff(K,Dq[i],Dq[j])

        # C(q,Dq)
        C = zeros(3,3)
        for i in range(0,3):
            for j in range(0,3):
                for k in range(0,3):
                    christoffel = Rational(1,2)*( diff(M[i,j],q[k]) + diff(M[i,k],q[j]) - diff(M[j,k],q[i]) )
                    C[i,j] += christoffel*Dq[k]
        # G(q)
        G = zeros(3,1)
        for i in range(0,len(G)):
            G[i,0] = diff(P,q[i])
        
        
        # Symbolic matrices
        self.sym_M = M
        self.sym_C = C
        self.sym_G = G

        
        # Symbolic state vectors
        self.sym_q = q
        self.sym_Dq = Dq
        
        # Symbolic energy expressions
        self.sym_K = K
        self.sym_P = P

        # creates the simulation dynamics
        # numeric dynamics are also created indirectly
        self.sim_M_C_G()
        self.sim_K_P()


    def relative_angles( self ):
        
        """
        Derivation of dynamics in relative angles        
        """
    	
        
        #--------- Symbolic Definitions -----------------------

        #define names of physical parameters as symbols
        l_1 = symbols('l_1')
        l_2 = symbols('l_2')
        l_3 = symbols('l_3')

        L_1 = symbols('L_1')
        L_2 = symbols('L_2')
        L_3 = symbols('L_3')

        m_1 = symbols('m_1')
        m_2 = symbols('m_2')
        m_3 = symbols('m_3')

        J_1 = symbols('J_1')
        J_2 = symbols('J_2')
        J_3 = symbols('J_3')

        g = symbols('g')


        #define generalized coordinates as functions of t
        #"Function" is a method from sympy module defining q_i to be functions of t
        t=symbols('t')
        q_1=Function('q_1')(t)
        q_2=Function('q_2')(t)
        q_3=Function('q_3')(t)

        #define dot(q_i) and ddot(q_i) to be symbols Dq_i and DDq_i
        Dq_1 = diff(q_1,t)
        Dq_2 = diff(q_2,t)
        Dq_3 = diff(q_3,t)
        DDq_1 = diff(q_1,t)
        DDq_2 = diff(q_2,t)
        DDq_3 = diff(q_3,t)

        # symbolic state vectors
        q = Matrix([[q_1],[q_2],[q_3]]) 
        Dq = Matrix([[Dq_1],[Dq_2],[Dq_3]])
        #-----------------------------------------------------


        #--------- Kinematic Relations -----------------------

        #Position 
        p_1=Matrix([
            [l_1*cos(q_1)],
            [l_1*sin(q_1)]
            ])
        p_2=Matrix([
            [L_1*cos(q_1)+l_2*cos(q_1+q_2)],
            [L_1*sin(q_1)+l_2*sin(q_1+q_2)]
            ])
        p_3=Matrix([
            [L_1*cos(q_1)+l_3*cos(q_1+q_2+q_3)],
            [L_1*sin(q_1)+l_3*sin(q_1+q_2+q_3)]
            ])

        #Linear Velocities
        v_1=p_1.diff(t)
        v_2=p_2.diff(t)
        v_3=p_3.diff(t)

        #------------------------------------------------------


        #--------- Kinetic and Potential Energy ---------------


        #Stance leg
        K_1 = (m_1/2)*(v_1[0]**2+v_1[1]**2) + (J_1/2)*(Dq_1**2)
        P_1 = m_1*g*p_1[1]

        #Torso
        K_2 = (m_2/2)*(v_2[0]**2+v_2[1]**2) + (J_2/2)*(Dq_1 + Dq_2)**2
        P_2 = m_2*g*p_2[1]

        #Swing leg
        K_3 = (m_3/2)*(v_3[0]**2 + v_3[1]**2) + (J_3/2)*(Dq_1 + Dq_2 + Dq_3)**2
        P_3 = m_3*g*p_3[1]

        #Collecting total kinetic and potential energy
        K = K_1 + K_2 + K_3
        P = P_1 + P_2 + P_3
        

        #-----------------------------------------------------

        #--------- Euler-Lagrange equations of motion --------


        # LHS-expressions
        LHS1 = diff(diff(K,Dq_1),t) - diff(K,q_1) + diff(P,q_1)
        LHS2 = diff(diff(K,Dq_2),t) - diff(K,q_2) + diff(P,q_2)
        LHS3 = diff(diff(K,Dq_3),t) - diff(K,q_3) + diff(P,q_3)
        
	# saving lhs of dynamical equations 
        self.LHS1 = LHS1
        self.LHS2 = LHS2
        self.LHS3 = LHS3
        

        #Find matrices M(q),C(q,Dq),G(q)

        #M(q)
        M = zeros(3,3)
        for i in range(0,3):
            for j in range(0,3):
                M[i,j] = diff(K,Dq[i],Dq[j])

        #print("::::::::: M(q) ::::::::::")
        #pprint(M)

        #C(q,Dq)
        C = zeros(3,3)
        for i in range(0,3):
            for j in range(0,3):
                for k in range(0,3):
                    christoffel = Rational(1,2)*( diff(M[i,j],q[k]) + diff(M[i,k],q[j]) - diff(M[j,k],q[i]) )
                    C[i,j] += christoffel*Dq[k]

        #print("\n\n:::::::: C(q,Dq) ::::::")
        #pprint(C)

        #G(q)
        G = zeros(3,1)
        for i in range(0,len(G)):
            G[i,0] = diff(P,q[i])
 
 
        # Symbolic matrices
        self.sym_M = M
        self.sym_C = C
        self.sym_G = G

        
        # Symbolic state vectors
        self.sym_q = q
        self.sym_Dq = Dq
        
        # Symbolic energy expressions
        self.sym_K = K
        self.sym_P = P

        # creates the simulation dynamics
        # numeric dynamics are also created indirectly
        self.sim_M_C_G()
        self.sim_K_P()       
        
    def num_M_C_G( self ):

        
        """
        creates numeric versions of the M,C,G matrices
        the matrices have symbolic state expressions,
        but with numeric physical paramaters
        """
        
        M, C, G = self.get_sym_M_C_G() 
        
        g, m1, m2, m3, L1, L2, L3, l1, l2, l3, J1, J2, J3 = self.get_data()          
        

        # convention: symbolic expressions have underscores
        # numeric does not. 
        M = M.subs([
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ])
       
        C = C.subs([
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ])         
	
        G = G.subs([
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ])
        

        self.num_M = M
        self.num_C = C
        self.num_G = G
     


    def sim_M_C_G( self ):
        
	"""
        creates a lambdify'd version of the M,C,G matrices for use in simulations
        """
        
        self.num_M_C_G()

        M, C, G = self.get_num_M_C_G()
        q, Dq = self.get_sym_q_Dq() 

        # lambdify does not work with symbolic drivaties
        # so symbolic dummies must be substitted in before lambdify

        dum_Dq_1 = symbols('Dq_1')
        dum_Dq_2 = symbols('Dq_2')
        dum_Dq_3 = symbols('Dq_3')
        dum_Dq = Matrix([[dum_Dq_1],[dum_Dq_2],[dum_Dq_3]])
       
        C=C.subs( [ (Dq[0],dum_Dq_1),(Dq[1],dum_Dq_2),(Dq[2],dum_Dq_3) ] )
        
        # for simulation C is essentially C(q,Dq)*Dq
        # so the matrix sim_C is really C(q,Dq)*Dq
        C = C*dum_Dq

        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']
        M = lambdify( (q,), M, modules=mat2array )
        C = lambdify( (q,dum_Dq), C, modules=mat2array )
        G = lambdify( (q,), G, modules=mat2array )
        self.sim_M = M
        self.sim_C = C
        self.sim_G = G

    def num_K_P( self  ):
        
	"""
        creates numeric versions of the expressions K,P
        the expressions have symbolic state expressions,
        but with numeric physical parameters
        """

        K, P = self.get_sym_K_P() 
        
        g, m1, m2, m3, L1, L2, L3, l1, l2, l3, J1, J2, J3 = self.get_data()          
        

        # convention: symbolic expressions have underscores
        # numeric does not. 
        K = K.subs([
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ])
       
        P = P.subs([
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ])
       
        self.num_K = K
        self.num_P = P

    def sim_K_P( self  ):
        
        """
        creates a lambdify'd version of the K,P expressions for use in simulations
        """
        
        self.num_K_P()
      
        K,P = self.get_num_K_P()

        q, Dq = self.get_sym_q_Dq() 

        # lambdify does not work with symbolic drivaties
        # so symbolic dummies must be substitted in before lambdify

        dum_Dq_1 = symbols('Dq_1')
        dum_Dq_2 = symbols('Dq_2')
        dum_Dq_3 = symbols('Dq_3')
        dum_Dq = Matrix([[dum_Dq_1],[dum_Dq_2],[dum_Dq_3]])
       
        K=K.subs( [ (Dq[0],dum_Dq_1),(Dq[1],dum_Dq_2),(Dq[2],dum_Dq_3) ] )
        
        K = lambdify( (q,dum_Dq), K, 'numpy')
        P = lambdify( (q,), P, 'numpy' )
        
        self.sim_K=K
        self.sim_P=P


    def load_Model_XML( self, xml_file_name ):
       
        """
        overrides the similiar function in the parent class
        loads the model data from xml and creates new dynamics
 
        """
 
        # getting the root element of the XML document 
        model = ET.parse( xml_file_name ).getroot()
       
        # passing it to the parser
        self.set_data_from_RootElement( model )
        g, m1, m2, m3, L1, L2, L3, l1, l2, l3, J1, J2, J3 = self.get_data()
        
        # saving new physical_parameter_subs as instance variable
        self.phys_params_subs = [
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ] 



    def reload_Model_XML( self, xml_file_name ):
       
 
        # getting the root element of the XML document 
        model = ET.parse( xml_file_name ).getroot()
       
        # passing it to the parser
        self.set_data_from_RootElement( model )
        g, m1, m2, m3, L1, L2, L3, l1, l2, l3, J1, J2, J3 = self.get_data()
        
        # saving new physical_parameter_subs as instance variable
        self.phys_params_subs = [
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ] 

        # setting up new dynamics
        self.sim_M_C_G()
        self.sim_K_P()




   
    def get_sym_M(self):
        return self.sym_M

    def get_sym_C(self):
        return self.sym_C
    
    def get_sym_G( self ):
        return self.sym_G

    def get_sym_M_C_G( self ):
        return self.sym_M, self.sym_C, self.sym_G

    def get_num_M_C_G( self ):
        return self.num_M, self.num_C, self.num_G
    
    def get_sym_q_Dq( self ):
        return self.sym_q, self.sym_Dq

    def get_LHS_1(self):
        return self.LHS1
    
    def get_LHS_2(self):
        return self.LHS2

    def get_LHS_3(self):
        return self.LHS3

    def get_sym_K( self ):
        return self.sym_K

    def get_sym_P( self ):
        return self.sym_P
	
    def get_sym_K_P( self ):
        return self.sym_K, self.sym_P
    
    def get_num_K( self ):
        return self.num_K

    def get_num_P( self ):
        return self.num_P
	
    def get_num_K_P( self ):
        return self.num_K, self.num_P   
	
    def get_sim_M_C_G( self ):
        return self.sim_M, self.sim_C, self.sim_G
    
    def get_sim_K_P( self ):
        return self.sim_K, self.sim_P

    def get_LHS( self ):

        lhs_1, lhs_2, lhs_3 = self.LHS1, self.LHS2, self.LHS3

        phys_params_subs = self.phys_params_subs

        return lhs_1.subs(phys_params_subs),lhs_2.subs(phys_params_subs),lhs_3.subs(phys_params_subs)

	
