# -*- coding: utf-8 -*-
"""
Created on Fri Nov 20 20:11:38 2015

@author: leoni_000
"""

import numpy as np
from numpy import pi
from scipy.integrate import odeint


def integrate_until_value( f, x_stop, x_0, dx_0, t ) :  
    
    """
    X_dX, T = integrate_until_value( f, x_stop, x_0, dx_0, t )
    
    function integrates a scalar 2nd order ODE given in the function "f" 
    from initial conditions x_0, dx_0
    along time sequence "t" 
    until "x" reaches value "x_stop"
    returned are the X_dX vector with np.array([x,dx]) values,
    where the last value is an approximation of the "x_stop" value
    which is iterated untill the termination threshold eps = 1.0e-10
    is reached as: np.abs( x_stop-x_best[0] ) > 1.0e-10 
    the returned time sequence "T" is of the same length as "X_dX" 
    of the time values from the "t" sequence except the last one, which 
    is an approximation of the crossing time 
    """
    
    # finding the very first crossing along the "t" sequence
    X_dX, T = integrate_until_value_crossing( f, x_stop, x_0, dx_0, t )
    
    x_dx_N   = X_dX[-1]  
    x_dx_N_1 = X_dX[-2]  
    
    t_N   = T[-1]
    t_N_1 = T[-2]

    # as a start of the loop ... it is actually not necessarily the best 
    # x_dx_N_1 can be a closer approximation .. 
    # but we cannot have a duplicate value at the end 
    x_best  = x_dx_N 
    t_best  = t_N
    # ... should actuqlly check here if the x_dx_N_1 is below the accuracy theshold  

    N_iter = 1
    
    while np.abs( x_stop-x_best[0] ) > 1.0e-10 : 
   
        if N_iter > 100 :   
            print( 'WARNING: couldnt reach the x_stop crossing within 100 iterations!!!' )
            X_dX[-1] = x_best
            T[-1]    = t_best
            return X_dX, T

        if (x_stop-x_dx_N[0])*(x_stop-x_dx_N_1[0]) < 0.0 :
            
            # average dx along the (N-1, N) time interval  
            dx_avrg = (x_dx_N[1]+x_dx_N_1[1])*0.5  
    
            # predicting the crossing time 
            # v = dx/dt => dt = dx/v
            t_cross = t_N_1 + (x_stop-x_dx_N_1[0]) / dx_avrg
            
            # compiling the new time sequence 
            T_n = [ t_N_1 ]
            dt_N = t_N - t_N_1
            if t_cross - dt_N*0.1 > t_N_1 :
                T_n.append( t_cross - dt_N*0.1 ) 
            T_n.append( t_cross )
            T_n.append( t_cross + dt_N*0.1 ) 
    
            # integrating the system along the new time sequence 
            # in the time interval around the predicted crossing time  
            X_dX_, T_ = integrate_until_value_crossing( f, x_stop, x_dx_N_1[0], x_dx_N_1[1], T_n )

            x_dx_N   = X_dX_[-1]  
            x_dx_N_1 = X_dX_[-2]  
            
            t_N   = T_[-1]
            t_N_1 = T_[-2]
            
            x_best = x_dx_N 
            t_best = t_N
            if np.abs( x_stop-x_best[0] ) > np.abs( x_stop-x_dx_N_1[0] ) : 
                x_best = x_dx_N_1 
                t_best = t_N_1
            
        else : 
            print( "WARNING: smth went wrong with the x_stop crossing iterations" )
            X_dX[-1] = x_best
            T[-1]    = t_best
            return X_dX, T

        N_iter +=1 
        
    X_dX[-1] = x_best # the best found so far  
    T[-1]    = t_best

    return X_dX, T

def integrate_until_value_crossing( f, x_stop, x_0, dx_0, t ) :  
    
    """
    X_dX, T = integrate_until_value_crossing( f, x_stop, x_0, dx_0, t )

    integrating the scalar 2nd order ODE given in the function "f" 
    from initial conditions x_0, dx_0
    along time sequence "t" 
    until "x" reaches value "x_stop"
    returned are the X_dX vector with np.array([x,dx]) values,
    where the last value is the first one after reaching the "x_stop"
    the returned time sequence "T" is of the same length as "X_dX" 
    of the time values from the "t" sequence
    
    the function is meant as a single step in the refinment 
    of the time moment at which ODE crosses "x_stop"  
    """
    
    x_dx_0 = np.array([ x_0, dx_0 ])

    X_dX = [ x_dx_0 ]  
    T = [ t[0] ]    
    
    x_dx = odeint(  f, x_dx_0, [ t[0], t[1] ]  )
    x_dx = x_dx[1]

    X_dX.append( x_dx )
    T.append( t[1] )
    
    i = 2
    while ( x_stop - x_dx_0[0] ) * ( x_stop - x_dx[0] ) > 0.0 and i<len(t) :
        x_dx_0 = x_dx 
        x_dx = odeint(  f, x_dx_0, [ t[i-1], t[i] ]  )
        x_dx = x_dx[1]
        X_dX.append( x_dx )
        T.append( t[i] )
        i += 1
    
    if ( x_stop - x_dx_0[0] ) * ( x_stop - x_dx[0] ) > 0.0 and i==len(t) :   
        print( 'WARNING: couldnt reach the x_stop crossing within the given time sequence!!!' )
        
    return X_dX, T 


def integrate_until_condition_GE0( f_rhs, f_stop, x_0, t ) :  
    
    """
    X, T = integrate_until_condition( f_rhs, f_stop, x_0, t )
    
    function integrates a 1st order ODE given by the function "f_rhs" 
    from initial conditions x_0
    along time sequence "t" 
    until "f_stop(x)" is more or equial 0.0 
    returned are the X vector with values,
    where the last value is an approximation of the "x_stop" value
    which is iterated untill the termination threshold eps = 1.0e-10
    the returned time sequence "T" is of the same length as "X" 
    of the time values from the "t" sequence except the last one, which 
    is an approximation of the crossing time 
    """
    
    if f_stop( x_0 ) < 0.0 :
        # need to rise a proper exception ... 
        print( 'ERROR: integrate_until_condition_GE0(...) the initial value of the stopping condition is below zero!' )
        return 
    
    # finding the very first crossing along the "t" sequence
    X, T = integrate_while_condition_GE0( f_rhs, f_stop, x_0, t )
    
    x_N   = X[-1]  
    x_N_1 = X[-2]  
    
    f_stop_N   = f_stop( x_N )   
    f_stop_N_1 = f_stop( x_N_1 )  
    
    if f_stop_N > 0.0 :
        print( 'WARNING: integrate_until_condition_GE0: cannot reach condition f_stop(x)<0.0 on the given time sequence' )
        return np.array(X), np.array(T)

    t_N   = T[-1]
    t_N_1 = T[-2]

    # as a start of the loop ... it is actually not necessarily the best 
    # x_dx_N_1 can be a closer approximation .. 
    # but we cannot have a duplicate value at the end 
    f_stop_best = f_stop_N
    t_best      = t_N
    x_best      = x_N 

    # N_iter = 1
    
    while np.abs( f_stop_best ) > 1.0e-12 : 

        # the last time step  
        dt_N = t_N - t_N_1
        
        # finding a linear approximation of the zero crossing time 
        # x_N / tN_cross == - x_N_1 / (dt_N-tN_cross) # t_N_1+tN_cross approx of the crossing time 
        # x_N * (dt_N-tN_cross) == - x_N_1*tN_cross
        # x_N * dt_N == (x_N - x_N_1)*tN_cross
        # x_N * dt_N / (x_N - x_N_1) == tN_cross
        tN_cross = f_stop_N * dt_N / (f_stop_N - f_stop_N_1)
        
        # compiling the new refined time sequence using the estimation of the crossing time  
        T_n = [ t_N_1 ]
        if tN_cross - dt_N*0.1 > 0.0 :
            T_n.append( t_N_1 + tN_cross - dt_N*0.1 ) 
        T_n.append( t_N_1 + tN_cross )
        T_n.append( t_N_1 + tN_cross + dt_N*0.1 )
        if tN_cross + dt_N*0.1 < dt_N :
            # ... apparently there are cases, when the linear approx is that bad
            # in case it is not true the last step will not happen 
            T_n.append( t_N )  

        # integrating the last interval again with better precicion 
        # attempting to "fork" the crossing moment 
        # using the linear approx of the f_stop(x)
        X_, T_ = integrate_while_condition_GE0( f_rhs, f_stop, x_N_1, T_n )

        x_N   = X_[-1]  
        x_N_1 = X_[-2]  
        
        f_stop_N   = f_stop( x_N )   
        f_stop_N_1 = f_stop( x_N_1 )  

        t_N   = T_[-1]
        t_N_1 = T_[-2]
        
        f_stop_best = f_stop_N 
        t_best      = t_N
        x_best      = x_N 
#        if np.abs( f_stop_best ) > np.abs( f_stop_N_1 ) : 
#            f_stop_best = f_stop_N_1 
#            t_best      = t_N_1
#            x_best      = x_N_1 
            
        # N_iter +=1
            
    X[-1] = x_best # the best found so far  
    T[-1] = t_best

    return np.array(X), np.array(T)
    

def integrate_while_condition_GE0( f_rhs, f_stop, x_0, t ) :  

    """
    X, T = integrate_while_condition( f_rhs, f_stop, x_0, t )

    integrating the scalar 1std order ODE defined by the RHS function "f_rhs" 
    from initial conditions x_0
    along time sequence "t" 
    while f_stop( x, t ) (... ???) is above zero   
    returned are the "X" vector with "x" values,
    where the last value is the first one after reaching the "f_stop" condition
    the returned time sequence "T" is of the same length as "X" 
    of the time values from the "t" sequence
    
    the function is meant as a single step in the refinment 
    of the time moment at which ODE crosses f_stop(...) == 0  
    """
    
    X = [ x_0 ]  
    T = [ t[0] ]    

    x_0_ = x_0 
    t_len = len(t)
    i = 1
    while True :  # emulating the do ... while loop 
        x = odeint(  f_rhs, x_0_, [ t[i-1], t[i] ]  ) # integrate another one 
        x_0_ = x[1]                                   # the final vector    
        X.append( x_0_ )                              # storing the values 
        T.append( t[i] )
        i += 1                                        # increasing the index
        if f_stop( x_0_ ) < 0.0 or i >= t_len : 
            break 
    
    return X, T 


def int_constant_step_2nd_order_ODE( rhs_fun, x_0, v_0, T ) : 

    """ a simple constant step solver for 2nd order ODE (mechanical equations)
    for Matlab/Python cross verification purposes 
    """
    
    x = np.zeros(  ( len(T), len(x_0) )  )
    v = np.zeros(  ( len(T), len(v_0) )  )

    dt = T[1] - T[0]

    x[0,:] = x_0
    v[0,:] = v_0

    for i in range(len(T)-1) : 

        v[i+1,:] = v[i,:] + rhs_fun( x[i,:], v[i,:], T[i] ) * dt 
    
        #	% a typical 1st order ODE solver
        #	% x(:,i+1) = x(:,i) + v(:,i) * dt 
    
        #	a version of midpoint
        #	same as: x(:,i+1) = x(:,i) + v(:,i) * dt + 0.5*RHS_fun( t(i), x(:,i), v(:,i) ) * dt^2; 
        # which is a higher order formula
        x[i+1,:] = x[i,:] + 0.5*( v[i+1,:] + v[i,:] ) * dt

    return x, v



def abs_to_rel(Q_DQ_abs):

    """
    converts absolute Q_DQ 6d vector into the relative conevtnion
    """
    
    q1_abs, q2_abs, q3_abs = Q_DQ_abs[0], Q_DQ_abs[1], Q_DQ_abs[2] 
    Dq1_abs, Dq2_abs, Dq3_abs = Q_DQ_abs[3], Q_DQ_abs[4], Q_DQ_abs[5] 

    q1_rel = q1_abs + pi/2
    q2_rel = q2_abs - q1_abs
    q3_rel = q3_abs - q2_abs + pi

    Dq1_rel = Dq1_abs 
    Dq2_rel = Dq2_abs - Dq1_abs
    Dq3_rel = Dq3_abs - Dq2_abs

    Q_DQ_rel = np.array([[q1_rel],[q2_rel],[q3_rel],[Dq1_rel],[Dq2_rel],[Dq3_rel]])

    return Q_DQ_rel

def rel_to_abs(Q_DQ_rel):
    
    """
    converts relative Q_DQ 6d vector into the absolute conevtnion
    """
    
    q1_abs, q2_abs, q3_abs = Q_DQ_rel[0], Q_DQ_rel[1], Q_DQ_rel[2] 
    Dq1_abs, Dq2_abs, Dq3_abs = Q_DQ_rel[3], Q_DQ_rel[4], Q_DQ_rel[5]

    q1_abs = q1_rel - pi/2
    q2_abs = q1_rel + q2_rel - pi/2
    q3_abs = q1_rel + q2_rel + q3_rel - (3/2)*pi

    Dq1_abs = Dq1_rel 
    Dq2_abs = Dq1_rel + Dq2_rel 
    Dq3_abs = Dq1_rel + Dq2_rel + Dq3_rel 

    Q_DQ_abs = np.array([[q1_rel],[q2_rel],[q3_rel],[Dq1_rel],[Dq2_rel],[Dq3_rel]])

    return Q_DQ_abs

