from biped_dynamics_v1 import Biped_Dynamics 
import numpy as np



class Biped( Biped_Dynamics ) :
    
    """ 
    The Main Biped class
    contains system state and Right Hand Side of the dynamical equations 
    as well as kinetic and potential energy functions in point and vectorized 
    forms ( point for a single calculation and vectorized for post-processing 
    ODE results )
    Conservation of angular momentum on impact of the swing leg is calculated 
    taking into account the left/right (swing/stance) leg reassignment
    on the impact
    """
    
    def __init__( self, convention, model_xml ) :
        
        # the parent class constructor
        super( Biped, self ).__init__( convention, model_xml )
        
        # default value) of state variables 
        self.q1  = 0.0
        self.q2  = 0.0
        self.q3  = 0.0
        self.dq1 = 0.0
        self.dq2 = 0.0
        self.dq3 = 0.0
    

    def RHS_3D_t( self, Q, dQ, t ) :
        return  self.RHS_3D( Q, dQ )
        
    def RHS_3D( self, Q, dQ ) :

        """
        function: ddQ = RHS_3D( Q, dQ )
        resolves the dd_q, which is the second derivative of the 
        state vector q ... according to the dynamic equations of the system 
        in relative coordinates:

        M(q) * dd_q + B( q, d_q ) * d_q  + G(q) = [ 0, 0, 0 ]^T
   
        returns ddQ np.array 
        """        
 
        M, C, G = self.get_sim_M_C_G()
       
	# resolve the RHS equation for 3D ddQ
        return np.hstack( np.linalg.solve( M(Q), -C(Q,dQ) - G(Q) ) ) 

    def RHS_6D_t( self, Q_dQ, t  ) :
        
        """
        function: ddQ = RHS_6D_t( Q_dQ, t )
        adds the silent time entry "t" to satisfy calling conditions of ODE 
        solvers ... otherwise identical to RHS_6D( self, Q_dQ ), 
        which is time independent         
        """        
        
        return self.RHS_6D( Q_dQ )
        
    def RHS_6D( self, Q_dQ ) :
        
        """
        function: ddQ = RHS_6D( Q_dQ )
        resolves the dd_q, which is the second derivative of the 
        state vector q ... according to the dynamic equations of the system 
        in relative coordinates:

        M(q) * dd_q + B( q, d_q ) * d_q  + G(q) = [ 0, 0, 0 ]^T
   
        the returned 6x1 np.array is dQ_ddQ for the dynamic system 
        written in block form as a 1st order nonlinear ODE 
        """        
 
        Q  = Q_dQ[0:3]
        dQ = Q_dQ[3:6]
        rhs_3D = self.RHS_3D( Q, dQ )

        return np.array(  [ dQ[0], dQ[1], dQ[2], rhs_3D[0], rhs_3D[1], rhs_3D[2] ]  )        
    


    def kinetic( self ) :

        """
        function: K = kinetic( )
        returns a scalar value K of kinetic energy for the current set of 
        coordinates set by set_State( Q_dQ ) method of the class, 
        where Q_dQ is a 6x1 np.array of coordinates and velocities 
        """

        return self.kinetic_vectorized( np.array([ self.q1, self.q2, self.q3, self.dq1, self.dq2, self.dq3 ]) )

    def kinetic_vectorized( self, Q_dQ ) :
    
        """
        function: K = kinetic_vectorized( Q_dQ )
        returns kinetic energy vector K (of lenght N) for a 6xN np.array of 
        coordinates and velocities
        """
        q = Q_dQ[:,0:3]
	Dq = Q_dQ[:,3:6]
        
        K = np.empty(len(Q_dQ))

        for t in range(0,len(Q_dQ)):
            K[t] = self.sim_K( q[t,:], Dq[t,:] ) 

        return K
    
 
    def potential( self ) :
        
        """
        function: P = potential( )
        returns a scalar value P of potential energy for the current set of 
        coordinates set by set_State( Q_dQ ) method of the class, 
        where Q_dQ is a 6x1 np.array of coordinates and velocities 
        """

        return self.potental_vectorized( np.array([ self.q1, self.q2, self.q3 ]) )

    def potential_vectorized( self, Q_dQ ) : 
        
        """
	function: P = potential_vectorized( self, Q_dQ )
        returns potential energy vector P (of lenght N) for a 6xN np.array of 
        coordinates and velocities (or for a 3xN np.array of coordinates only)
        """
        q = Q_dQ[:,0:3]
        
        P = np.empty(len(Q_dQ))

        for t in range(0,len(Q_dQ)):
            P[t] =  self.sim_P( q[t,:] ) 

        return P
    
        
# ----------------------------------------------------------------
