import sys
sys.path.append("../biped/")
sys.path.append("../model_parameters_xml/")

from biped_data import Biped_Data 
from sympy import *
import sympy
import numpy as np



class Q_Impact_Map( Biped_Data ):

    def __init__( self, convention, xml_file_name ):
        
        super( Q_Impact_Map, self ).__init__()
        
        self.convention = convention
        self.load_Model_XML(xml_file_name)
        self.derive_q_impact_map()
        


    def derive_q_impact_map( self ):
        
        
        
        #--------- Symbolic Definitions -----------------------

        #define names of physical parameters as symbols
        l_1 = symbols('l_1')
        l_2 = symbols('l_2')
        l_3 = symbols('l_3')

        L_1 = symbols('L_1')
        L_2 = symbols('L_2')
        L_3 = symbols('L_3')

        m_1 = symbols('m_1')
        m_2 = symbols('m_2')
        m_3 = symbols('m_3')

        J_1 = symbols('J_1')
        J_2 = symbols('J_2')
        J_3 = symbols('J_3')

        g = symbols('g')


        # define generalized coordinates as functions of t
        # "Function" is a method from sympy module defining q_i to be functions of t
        # simply q_i, Dq_i refers to values pre-impact (-)

        t=symbols('t')
        q_1 = Function('q_1')(t)
        q_2 = Function('q_2')(t)
        q_3 = Function('q_3')(t)

        #define dot(q_i) and ddot(q_i) to be symbols Dq_i
        Dq_1 = diff(q_1,t)
        Dq_2 = diff(q_2,t)
        Dq_3 = diff(q_3,t)


        # dummies for subs and lambdify

        q_1_minus = Function('q_1_minus')(t)
        q_2_minus = Function('q_2_minus')(t)
        q_3_minus = Function('q_3_minus')(t)

        Dq_1_minus = symbols('Dq_1_minus')
        Dq_2_minus = symbols('Dq_2_minus')
        Dq_3_minus = symbols('Dq_3_minus')

        

        # symbolic state vectors
        q = Matrix([[q_1],[q_2],[q_3]]) 
        Dq = Matrix([[Dq_1],[Dq_2],[Dq_3]])
        
        q_minus=Matrix([[q_1_minus],[q_2_minus],[q_3_minus]])
        Dq_minus = Matrix([[Dq_1_minus],[Dq_2_minus],[Dq_3_minus]])

        
        
        #-----------------------------------------------------


        """
        
        Frame minus has  origo at 
        the end of stance leg pre-impact.

        Frame plus has origo at 
        the end of (new) stance leg post-impact

        Angular momentum is conserved about
        two points in both frames
        
        o_2 (hip)
        
        o_3 (end of swing leg pre-impact)/
            (end of stance leg post-impact)

        """
        
        if( self.convention == "absolute" ):
            

            # position vectors for
            # links' CoM referenced o_1
            p_1_o1 = Matrix([
                [-l_1*sin(q_1)],
                [l_1*cos(q_1)],
                [0]
                ])

            p_2_o1 = Matrix([
                [-L_1*sin(q_1)-l_2*sin(q_2)],
                [L_1*cos(q_1)+l_2*cos(q_2)],
                [0]          
                ])

            p_3_o1 = Matrix([
                [-L_1*sin(q_1)+l_3*sin(q_3)],
                [L_1*cos(q_1)-l_3*cos(q_3)],
                [0]          
                ])

            o2 = Matrix([
                [-L_1*sin(q_1)],
                [L_1*cos(q_1)],
                [0]
                ])

      
            o3 = Matrix([
                [-L_1*sin(q_1)+L_3*sin(q_3)],
                [L_1*cos(q_1)-L_3*cos(q_3)],
                [0]          
                ])


            q_1_plus = - q_1_minus
            q_2_plus = q_2_minus
            q_3_plus = - q_3_minus

            q_plus = Matrix([
               [q_1_plus],
               [q_2_plus],
               [q_3_plus]
               ])
            
            
            # angular momentum about links
            # own CoM ref inertial frame

            J1_Dq1 = J_1*Dq_1
            J2_Dq2 = J_2*Dq_2
            J3_Dq3 = J_3*Dq_3



        if( self.convention == "relative" ):

            p_1_o1 = Matrix([
                [l_1*cos(q_1)],
                [l_1*sin(q_1)],
                [0]
                ])
            p_2_o1 = Matrix([
                [L_1*cos(q_1)+l_2*cos(q_1+q_2)],
                [L_1*sin(q_1)+l_2*sin(q_1+q_2)],
                [0]
                ])
            p_3_o1 = Matrix([
                [L_1*cos(q_1)+l_3*cos(q_1+q_2+q_3)],
                [L_1*sin(q_1)+l_3*sin(q_1+q_2+q_3)],
                [0]
                ])

            o2 = Matrix([
                [L_1*cos(q_1)],
                [L_1*sin(q_1)],
                [0]
                ])

            o3 = Matrix([
                [L_1*cos(q_1)+L_3*cos(q_1+q_2+q_3)],
                [L_1*sin(q_1)+L_3*sin(q_1+q_2+q_3)],
                [0]
                ])

            
            q_23 = np.pi - 2.0*q_1_minus

            q_1_plus = np.pi - q_1_minus
            q_2_plus = -(q_23 - q_2_minus)
            q_3_plus = q_3_minus - q_23

            q_plus = Matrix([
               [q_1_plus],
               [q_2_plus],
               [q_3_plus]
               ])
            

            # angular momentum about links
            # own CoM ref inertial frame

            J1_Dq1 = J_1*Dq_1
            J2_Dq2 = J_2*(Dq_1 + Dq_2)
            J3_Dq3 = J_3*(Dq_1 + Dq_2 + Dq_3)

            
        
        # position vectors for
        # link's CoM referenced o_2

        p_1_o2 = p_1_o1 - o2
        p_2_o2 = p_2_o1 - o2
        p_3_o2 = p_3_o1 - o2

        # position vectors for
        # link's CoM referenced o_3

        p_1_o3 = p_1_o1 - o3
        p_2_o3 = p_2_o1 - o3
        p_3_o3 = p_3_o1 - o3
        

        # velocties
        v_1_o1 = p_1_o1.diff(t)
        v_2_o1 = p_2_o1.diff(t)
        v_3_o1 = p_3_o1.diff(t)

        v_1_o2 = p_1_o2.diff(t)
        v_2_o2 = p_2_o2.diff(t)
        v_3_o2 = p_3_o2.diff(t)

        v_1_o3 = p_1_o3.diff(t)
        v_2_o3 = p_2_o3.diff(t)
        v_3_o3 = p_3_o3.diff(t)



        #:::::::::::::::BEFORE IMPACT:::::::::::::::::::

        # Angular momentum of entire robot
        # about swing leg
        RHS_11 = p_1_o3.cross(m_1*v_1_o1)[2] + J1_Dq1
        RHS_12 = p_2_o3.cross(m_2*v_2_o1)[2] + J2_Dq2
        RHS_13 = p_3_o3.cross(m_3*v_3_o1)[2] + J3_Dq3
        RHS_3 = RHS_11 + RHS_12 + RHS_13
        
        # about hip
        RHS_1 = p_1_o2.cross(m_1*v_1_o1)[2] + J1_Dq1
        RHS_2 = p_2_o2.cross(m_2*v_2_o1)[2] + J2_Dq2

        
        
        #:::::::::::::::AFTER IMPACT::::::::::::::::::::


        # Angular momentum of entire robot
        # about swing leg
        LHS_31 = p_1_o1.cross(m_1*v_1_o1)[2] + J1_Dq1
        LHS_32 = p_2_o1.cross(m_2*v_2_o1)[2] + J2_Dq2
        LHS_33 = p_3_o1.cross(m_3*v_3_o1)[2] + J3_Dq3
        LHS_3 = LHS_31 + LHS_32 + LHS_33

        # about hip
        LHS_1 = p_3_o2.cross(m_3*v_3_o1)[2] + J3_Dq3
        LHS_2 = p_2_o2.cross(m_2*v_2_o1)[2] + J2_Dq2

        #:::::::::::::::::::::::::::::::::::::::::::::::


        self.LHS_1 = LHS_1
        self.LHS_2 = LHS_2
        self.LHS_3 = LHS_3

        self.RHS_1 = RHS_1
        self.RHS_2 = RHS_2
        self.RHS_3 = RHS_3

        LHS_1 = LHS_1.expand()
        LHS_2 = LHS_2.expand()
        LHS_3 = LHS_3.expand()

        RHS_1 = RHS_1.expand()
        RHS_2 = RHS_2.expand()
        RHS_3 = RHS_3.expand()
        
        #Formulating impact matrices and vectors
        A = Matrix([
            [LHS_1.coeff(Dq_1),LHS_1.coeff(Dq_2),LHS_1.coeff(Dq_3)],
            [LHS_2.coeff(Dq_1),LHS_2.coeff(Dq_2),LHS_2.coeff(Dq_3)],
            [LHS_3.coeff(Dq_1),LHS_3.coeff(Dq_2),LHS_3.coeff(Dq_3)]
            ])

        B = Matrix([
            [RHS_1.coeff(Dq_1),RHS_1.coeff(Dq_2),RHS_1.coeff(Dq_3)],
            [RHS_2.coeff(Dq_1),RHS_2.coeff(Dq_2),RHS_2.coeff(Dq_3)],
            [RHS_3.coeff(Dq_1),RHS_3.coeff(Dq_2),RHS_3.coeff(Dq_3)]
            ])

        self.sym_A = A
        self.sym_B = B

        LHS = Matrix([
            [LHS_1],
            [LHS_2],
            [LHS_3]
            ])
        
        RHS = Matrix([
            [RHS_1],
            [RHS_2],
            [RHS_3]
            ])

        # setting up the linear impact problem
        # substituting and lambdifying A_matrix and B-vecctor for resolving impact
        g, m1, m2, m3, L1, L2, L3, l1, l2, l3, J1, J2, J3 = self.get_data()

        A = A.subs([
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (q_1,q_1_plus),(q_2,q_2_plus),(q_3,q_3_plus),
            (symbols('g'),g)
            ])

        B = RHS.subs([
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (Dq_1,Dq_1_minus),(Dq_2,Dq_2_minus),(Dq_3,Dq_3_minus),
            (symbols('g'),g)
            ])

        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']

        self.A = lambdify( (q_minus,), A, modules=mat2array )
        self.B = lambdify( (q,Dq_minus), B, modules=mat2array )
        self.q_plus = lambdify( (q_minus,), q_plus, modules=mat2array )





    def impact( self, Q_DQ_minus ):

        """
        takes the 6d Q_DQ state vector pre-impact
        and returns 6d Q_DQ state vector post-impact
        """
        
        q_minus = Q_DQ_minus[0:3]
        Dq_minus = Q_DQ_minus[3:6]
        
        Q_plus = self.q_plus(q_minus)
        
        DQ_plus = np.linalg.solve(self.A(q_minus),self.B(q_minus,Dq_minus))

        return np.concatenate((Q_plus,DQ_plus))



            
            
    def get_A_matrix( self ):

        A = self.sym_A 
        
        g, m1, m2, m3, L1, L2, L3, l1, l2, l3, J1, J2, J3 = self.get_data()

        phys_params_subs = [
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ]

        return A.subs( phys_params_subs )




    def get_B_matrix( self ):
        
        B = self.sym_B
        
        g, m1, m2, m3, L1, L2, L3, l1, l2, l3, J1, J2, J3 = self.get_data()

        phys_params_subs = [
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ]

        return B.subs( phys_params_subs )



    
    def get_sym_LHS( self ):
        
        """
        returns a sympy matrix of symbolic lhs impact equations
        the current physical parameters ar subbed in
        """


        LHS = Matrix([
            [self.LHS_1],
            [self.LHS_2],
            [self.LHS_3]
            ])
       
        g, m1, m2, m3, L1, L2, L3, l1, l2, l3, J1, J2, J3 = self.get_data()

        phys_params_subs = [
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ]


        return LHS.subs( phys_params_subs )




    def get_sym_RHS( self ):
        
        
        """
        returns a sympy matrix of symbolic rhs impact equations
        the current physical parameters ar subbed in
        """
       
        RHS = Matrix([
            [self.RHS_1],
            [self.RHS_2],
            [self.RHS_3]
            ])
        
        g, m1, m2, m3, L1, L2, L3, l1, l2, l3, J1, J2, J3 = self.get_data()

        phys_params_subs = [
            (symbols('m_1'),m1),(symbols('m_2'),m2),(symbols('m_3'),m3),
            (symbols('L_1'),L1),(symbols('L_2'),L2),(symbols('L_3'),L3),
            (symbols('l_1'),l1),(symbols('l_2'),l2),(symbols('l_3'),l3),
            (symbols('J_1'),J1),(symbols('J_2'),J2),(symbols('J_3'),J3),
            (symbols('g'),g)
            ]

        return RHS.subs( phys_params_subs )

