import numpy as np
from scipy.integrate import odeint
import biped_utils



class Motion_Generator( object ):
    
    """ 
    generate q-space trajectories and torques
    from nominal theta-trajectories
    """

    def __init__( self, abg, vhc ):
        
        self.abg = abg
        self.vhc = vhc 

    

    def abg0_sim(self, theta_theta_d_0, T):

        """
        integrates abg0-dynamics along time-vector
        returns array theta_theta_d
        """

        theta_theta_d = odeint(self.abg.rhs_0_t, theta_theta_d_0, T)

        return theta_theta_d
        


    def abg0_sim_until_value(self, theta_theta_d_0, theta_stop, t ):

        """
        integrates the abg zero dynamics until theta_stop is reached
        """

        theta_0 = theta_theta_d_0[0]
        theta_d_0 = theta_theta_d_0[1]

        theta_theta_d, T = biped_utils.integrate_until_value(self.abg.rhs_0_t, theta_stop,  theta_0, theta_d_0, t)

        self.theta_theta_d = np.array(theta_theta_d)
        self.T = np.array(T)
        

        self.generate_torques()
        self.generate_Q_Qd()

        return np.array(theta_theta_d), np.array(T)

    
    def generate_Q_Qd( self ):

        """
        generates trajectories [q_i;q_i_d]
        from theta-trajectories
        """

        theta_theta_d = self.theta_theta_d

        phi1, phi2, phi3 = self.vhc.get_func_vhc()
        Dphi1, Dphi2, Dphi3 = self.vhc.get_func_Dphi()

        P1, P2, P3 = self.vhc.get_parameterization()

        theta = theta_theta_d[:,0]
        theta_d = theta_theta_d[:,1]

        q1, q2, q3, q1_d, q2_d, q3_d = [],[],[],[],[],[]
        
        for t in range(0,len(theta)):
            
            theta = theta_theta_d[t,0]
            theta_d = theta_theta_d[t,1]

            q1.append(phi1(theta,P1))
            q2.append(phi2(theta,P2))
            q3.append(phi3(theta,P3))
            q1_d.append(Dphi1(theta, P1)*theta_d) 
            q2_d.append(Dphi2(theta, P2)*theta_d) 
            q3_d.append(Dphi3(theta, P3)*theta_d) 

        Q_Qd = np.array([q1, q2, q3, q1_d, q2_d, q3_d]).T

        self.Q_Qd = Q_Qd

    
    def generate_torques( self ):

        """
        generates torques u_2 and u_3
        from theta-trajectories
        """

        theta_theta_d = self.theta_theta_d

        theta = theta_theta_d[:,0]
        theta_d = theta_theta_d[:,1]

        tau2, tau3 = self.abg.get_tau_sim()
        
        
        t2, t3 = [],[]
        
        for t in range(0,len(theta)):
        
            theta = theta_theta_d[t,0]
            theta_d = theta_theta_d[t,1]

            t2.append(tau2(theta, theta_d)) 
            t3.append(tau3(theta, theta_d)) 
        
        tau = np.array([t2, t3]).T

        self.tau = tau



    def get_Q_Qd( self ):

        return self.Q_Qd 

    def get_torques( self ):

        return self.tau

    def get_theta_traj( self ):

        return theta_theta_d, T
















