import sys
sys.path.append("../q_space_dynamics/")

from biped_data import Biped_Data 
from sympy import *
import numpy as np
from impact_map import Q_Impact_Map



class Theta_Impact_Map( Biped_Data ):

    
    def __init__( self, convention, xml_file_name, vhc, theta_0, theta_t ):
        
        
        super( Theta_Impact_Map, self ).__init__()

        self.theta_0 = theta_0
        self.theta_t = theta_t
        
        self.load_Model_XML(xml_file_name)
        self.vhc = vhc
        self.q_impact_map = Q_Impact_Map( convention, xml_file_name  )
        
        self.derive_theta_impact_map( convention )
	self.derive_symbolic_factors()
        self.lambdify_factors()


    
    def derive_theta_impact_map( self, convention ):
        
        
        # obtain imapact dynamics from q_impact_map
        q_impact_map = self.q_impact_map
        
	A = q_impact_map.get_A_matrix()
        B = q_impact_map.get_B_matrix()

        # obtain vhcs
        phi_1, phi_2, phi_3 = self.vhc.get_sym_vhcs()
        Dphi_1, Dphi_2, Dphi_3 = self.vhc.get_sym_Dphi()
        
        
        # ---------------------------symbolic  definitions -------------------------
        
        t = symbols('t')
        theta = Function('theta')(t)
        theta_d = diff(theta, t)

        q_1 = Function('q_1')(t)
        q_2 = Function('q_2')(t)
        q_3 = Function('q_3')(t)

        q_1_d = diff(q_1,t)
        q_2_d = diff(q_2,t)
        q_3_d = diff(q_3,t)

	Dphi_sym = Matrix([
		[Dphi_1],
		[Dphi_2],
		[Dphi_3]
		])
        
        if ( convention == "relative" ):
            
            phi_1_plus = np.pi - phi_1
            phi_2_plus = 2.0*phi_1 + phi_2 - np.pi
            phi_3_plus = phi_3 + 2.0*phi_1 - np.pi


        elif ( convention == "absolute" ):
            
            phi_1_plus = - phi_1
            phi_2_plus = phi_2
            phi_3_plus = - phi_3

        self.phi_1_plus, self.phi_2_plus, self.phi_3_plus = phi_1_plus, phi_2_plus, phi_3_plus
        
        self.phi_plus = Matrix([
            [phi_1_plus],
            [phi_2_plus],
            [phi_3_plus]
            ])
        

        # -----------------------------------------------------------------------------
        
        # substituting in vhcs to obtain imapct dynamics in theta-space

        vhc_subs_1 = [ (q_1, phi_1_plus),(q_2, phi_2_plus),(q_3, phi_3_plus) ]
        vhc_subs_3 = [ (q_1, phi_1),(q_2, phi_2),(q_3, phi_3) ]
        
        Q = A.subs( vhc_subs_1 )
        P = B.subs( vhc_subs_3 )

        self.Q_sym = Q
        self.P_sym = P
        self.Dphi_sym = Dphi_sym
        
        # vhc parameters
        vhc_parameters = self.vhc.get_full_parameter_list()

        P_1, P_2, P_3 = self.vhc.get_parameter_symbols()

        # deinfe lambdifying
        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']


        self.phi1_plus = lambdify( (theta, vhc_parameters), phi_1_plus, 'numpy' )
        self.phi2_plus = lambdify( (theta, vhc_parameters), phi_2_plus, 'numpy' )
        self.phi3_plus = lambdify( (theta, vhc_parameters), phi_3_plus, 'numpy' )

	#self.Dphi = lambdify( (theta, vhc_parameters), Dphi_sym, modules = mat2array)

        self.Q = lambdify( (theta, vhc_parameters), Q, modules = mat2array )
        self.P = lambdify( (theta, vhc_parameters), P, modules = mat2array )


    
    
    def impact(self, theta_d_t):
        
        theta_0, theta_t = self.theta_0, self.theta_t
        vhc_parameters = self.vhc.get_full_parameterization_list()
        
        Dphi_plus = self.vhc.Dphi(theta_0, vhc_parameters)
        Dphi_minus = self.vhc.Dphi(theta_t, vhc_parameters)

        Q = self.Q( theta_t, vhc_parameters )
        P = self.P( theta_t, vhc_parameters )
        
        LHS = Q.dot(Dphi_plus)
        RHS = P.dot(Dphi_minus)

        k1 = RHS[0]/LHS[0]
        k2 = RHS[1]/LHS[1]
        k3 = RHS[2]/LHS[2]

        K = np.array([k1, k2, k3])
        theta_d_0 = np.array([ k1*theta_d_t, k2*theta_d_t, k3*theta_d_t ])
        
        return theta_d_0, K           				
    
    
    def get_factors( self ):


        theta_0, theta_t = self.theta_0, self.theta_t
        vhc_parameters = self.vhc.get_full_parameterization_list()
        
        Dphi_plus = self.vhc.Dphi(theta_0, vhc_parameters)
        Dphi_minus = self.vhc.Dphi(theta_t, vhc_parameters)

        Q = self.Q( theta_t, vhc_parameters )
        P = self.P( theta_t, vhc_parameters )
        
        LHS = Q.dot(Dphi_plus)
        RHS = P.dot(Dphi_minus)

        k1 = RHS[0]/LHS[0]
        k2 = RHS[1]/LHS[1]
        k3 = RHS[2]/LHS[2]

        return float(k1), float(k2), float(k3)
    
    
    
    def derive_symbolic_factors( self ):

        theta_0, theta_t = self.theta_0, self.theta_t

        Q = self.Q_sym
        P = self.P_sym

        t = symbols('t')
        theta = Function('theta')(t)

        Dphi = self.Dphi_sym
        
        Dphi_plus = Dphi.subs(theta,theta_0)
        Dphi_minus = Dphi.subs(theta,theta_t)

        Q = Q.subs(theta, theta_t)
        P = P.subs(theta, theta_t)
        
        LHS = Q.dot(Dphi_plus)
        RHS = P.dot(Dphi_minus)
        
        self.k_1 = RHS[0]/LHS[0]
        self.k_2 = RHS[1]/LHS[1]
        self.k_3 = RHS[2]/LHS[2]

    
    def derive_factor_grad( self ):

        t = symbols('t')
        theta = Function('theta')(t)

        k_1 = self.k_1
        k_2 = self.k_2
        k_3 = self.k_3
        
        vhc_parameters = self.vhc.get_full_parameter_list()

        grad_1 = zeros(1,len(vhc_parameters))
        grad_2 = zeros(1,len(vhc_parameters))
        grad_3 = zeros(1,len(vhc_parameters))

        parameter_count = 0
        for p in vhc_parameters:
            
            grad_1[parameter_count] = diff(k_1,p) 
            grad_2[parameter_count] = diff(k_2,p) 
            grad_3[parameter_count] = diff(k_3,p)
            parameter_count += 1
        
        self.k_1_grad = grad_1
        self.k_2_grad = grad_2
        self.k_3_grad = grad_3

        # define lambdifying
        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']

        self.k1_grad = lambdify( (vhc_parameters,), grad_1, modules = mat2array )
        self.k2_grad = lambdify( (vhc_parameters,), grad_2, modules = mat2array )
        self.k3_grad = lambdify( (vhc_parameters,), grad_3, modules = mat2array )
    
        
    
    def lambdify_factors( self ):

        
        P = self.vhc.get_full_parameter_list()
        
        k_1 = self.k_1  
        k_2 = self.k_2  
        k_3 = self.k_3

        self.k1 = lambdify( (P,), k_1, 'numpy' ) 
        self.k2 = lambdify( (P,), k_2, 'numpy' ) 
        self.k3 = lambdify( (P,), k_3, 'numpy' ) 




    def get_sym_factors( self ):
        
        return self.k_1, self.k_2, self.k_3

    def get_sym_factor_grads( self ):
        
        return self.k_1, self.k_2, self.k_3

    def get_func_factor_grads( self ):
        
        return self.k1_grad, self.k2_grad, self.k3_grad

    def get_func_factors( self ):

        return self.k1, self.k2, self.k3
    
    def get_func_phi_plus( self ):

        return self.phi1_plus, self.phi2_plus, self.phi3_plus

    def get_sym_phi_plus( self ):

        return self.phi_1_plus, self.phi_2_plus, self.phi_3_plus
    





