import sys
sys.path.append("../q_space_dynamics/")

from sympy import *
from biped_dynamics_v1 import Biped_Dynamics


import random
import scipy.integrate
import numpy as np
import collections
import time
import lxml.etree as ET





class VHC( object ):

    """
    takes symbolic vhcs as argument and creates lambdified 
    versions with parameters as args

    """
    def __init__( self, phi_1, phi_2, phi_3 ):

        t = symbols('t')
        theta = Function('theta')(t)
        
        self.phi_1, self.phi_2, self.phi_3 = phi_1, phi_2, phi_3

        self.phi_1_d = diff(phi_1,t)
        self.phi_2_d = diff(phi_2,t)
        self.phi_3_d = diff(phi_3,t)

        self.phi_1_dd = diff(phi_1,t,t)
        self.phi_2_dd = diff(phi_2,t,t)
        self.phi_3_dd = diff(phi_3,t,t)

        self.Dphi_1 = diff(phi_1,theta)
        self.Dphi_2 = diff(phi_2,theta)
        self.Dphi_3 = diff(phi_3,theta)

	self.Dphi_sym = Matrix([
		[self.Dphi_1],
		[self.Dphi_2],
		[self.Dphi_3]
		])

        self.DDphi_1 = diff(phi_1,theta,theta)
        self.DDphi_2 = diff(phi_2,theta,theta) 
        self.DDphi_3 = diff(phi_3,theta,theta)

        self.parameter_handler()
        
        #lambdify expressions
        self.alt_lambdify_phi()
        self.lambdify_phi()
        self.lambdify_Dphi()
        


    def parameter_handler( self ):
        

        t = symbols('t')
       
        phi_1, phi_2, phi_3 = self.get_sym_vhcs() 
	
        # getting sypmpy symbols in a set
        # as sympy symbols
        P_1 = phi_1.free_symbols
        P_2 = phi_2.free_symbols
        P_3 = phi_3.free_symbols

        P_1.remove(t)
        P_2.remove(t)
        P_3.remove(t)
        
        # create listof strings 
        # corresponding to the sympy symbols of the vhcs 
        P_1_str = map(str, list(P_1))
        P_2_str = map(str, list(P_2))
        P_3_str = map(str, list(P_3))

	#create 
        
        # creating an unordered dictionary
        # relating parameters as strings (keys) to sympy symbols (values)
        P_1_str_P_1 = dict(zip(P_1_str,P_1)) 
        P_2_str_P_2 = dict(zip(P_2_str,P_2)) 
        P_3_str_P_3 = dict(zip(P_3_str,P_3)) 
        
        # creating ordered dictionary
        # of (strings:sympy symbols)
        self.P_1_str_P_1 = collections.OrderedDict(sorted(P_1_str_P_1.items(), key=lambda t: t[0]))
        self.P_2_str_P_2 = collections.OrderedDict(sorted(P_2_str_P_2.items(), key=lambda t: t[0]))
        self.P_3_str_P_3 = collections.OrderedDict(sorted(P_3_str_P_3.items(), key=lambda t: t[0]))
        
        # saving the list of sympy symbols as instance variables
        
	self.P_1, self.P_2, self.P_3 = self.get_parameter_symbols()



    

    def lambdify_phi( self ):

        """"
        creates functions out of the 
        symbolic phi-expressions
        phi_i(theta,P_i)
        """
        
        t = symbols('t')
        theta = Function('theta')(t)

        phi_1, phi_2, phi_3 = self.get_sym_vhcs()

        P_1, P_2, P_3 = self.get_parameter_symbols() 

        self.phi1 = lambdify( (theta,P_1), phi_1, 'numpy' )
        self.phi2 = lambdify( (theta,P_2), phi_2, 'numpy' )
        self.phi3 = lambdify( (theta,P_3), phi_3, 'numpy' )


    def lambdify_Dphi( self ):
        
        t = symbols('t')
        theta = Function('theta')(t)

        Dphi_1, Dphi_2, Dphi_3 = self.Dphi_1, self.Dphi_2, self.Dphi_3

        P_1, P_2, P_3 = self.get_parameter_symbols()

        vhc_parameters = self.get_full_parameter_list()

        self.Dphi1 = lambdify( (theta,P_1), Dphi_1, 'numpy' )
        self.Dphi2 = lambdify( (theta,P_2), Dphi_2, 'numpy' )
        self.Dphi3 = lambdify( (theta,P_3), Dphi_3, 'numpy' )

        mat2array = [{'ImmutableMatrix': np.array}, 'numpy']
	self.Dphi = lambdify( (theta, vhc_parameters), self.Dphi_sym, modules = mat2array)



    def alt_lambdify_phi( self ):

        """"
        creates functions out of the 
        symbolic phi-expressions.
        phi_i(theta,P)
        similar to lambdify_phi, but
        functions here take in the 
        whole P-vec
        """
        t = symbols('t')
        theta = Function('theta')(t)

        P = self.get_full_parameter_list()

        phi_1, phi_2, phi_3 = self.get_sym_vhcs()

        self.alt_phi1 = lambdify( (theta,P), phi_1, 'numpy' )
        self.alt_phi2 = lambdify( (theta,P), phi_2, 'numpy' )
        self.alt_phi3 = lambdify( (theta,P), phi_3, 'numpy' )



    def create_parameter_dicts( self ):

        P_1, P_2, P_3 = self.get_parameter_strings()
        P1, P2, P3 = self.get_parameterization()
        
        self.P_1_P1 = collections.OrderedDict(zip(P_1,P1)) 
        self.P_2_P2 = collections.OrderedDict(zip(P_2,P2))
        self.P_3_P3 = collections.OrderedDict(zip(P_3,P3))


    
    
    def random_parameterization( self, tup1, tup2, tup3 ):

        """
        random parameterization function
        written for absolute coordinates
        """

        P1, P2, P3 = [],[],[]
        n1, n2, n3 = self.get_number_of_parameters()

        for i in range(0,n1):
            P1.append( random.uniform( tup1[i][0], tup1[i][1] ))

        for i in range(0,n2):
            P2.append( random.uniform( tup2[i][0], tup2[i][1] ) )
	
	P3.append(-P1[0])
        for i in range(1,n3):
            P3.append( random.uniform( tup3[i][0], tup3[i][1] ) )        

        self.P1, self.P2, self.P3 = np.array(P1), np.array(P2), np.array(P3)
        
        self.Parameterization = np.concatenate((P1,P2,P3))
    
    
    def rel_random_parameterization( self, tup1, tup2, tup3 ):
        
        """
        random parameterization function
        written for relative coordinates
        """

        P1, P2, P3 = [],[],[]
        n1, n2, n3 = self.get_number_of_parameters()


        for i in range(0,n1):
            P1.append( random.uniform( tup1[i][0], tup1[i][1] ))

        for i in range(0,n2):
            P2.append( random.uniform( tup2[i][0], tup2[i][1] ) )
	
	P3.append(2*np.pi - 2*P1[0] - P2[0])	
        for i in range(1,n3):
            P3.append( random.uniform( tup3[i][0], tup3[i][1] ) )        

        self.P1, self.P2, self.P3 = np.array(P1), np.array(P2), np.array(P3)
        
        self.Parameterization = np.concatenate((P1,P2,P3))



    def random_reparameterization( self, rand_range, lower, upper, roam ):
        
        
        # get the current parameterization
        P1, P2, P3 = self.get_parameterization()
        len_1, len_2, len_3 = self.get_number_of_parameters()

        
        plus = abs(rand_range)
        minus = -abs(rand_range)


        for i in range(0,len_1):

            p = P1[i]
            p = p + random.uniform(minus, plus)
            if p < lower and not roam :
                P1[i] = lower
            elif p > upper and not roam:
                P1[i] = upper
            else:
                P1[i] = p

        for i in range(0,len_2):
            
            p = P2[i]
            p = p + random.uniform(minus, plus)
            if p < lower and not roam :
                P2[i] = lower
            elif p > upper and not roam:
                P2[i] = upper
            else:
                P2[i] = p

        for i in range(0,len_3):
            
            p = P3[i]
            p = p + random.uniform(minus, plus)
            if p < lower and not roam:
                P3[i] = lower
            elif p > upper and not roam:
                P3[i] = upper
            else:
                P3[i] = p
       
        self.P1, self.P2, self.P3 = P1, P2, P3
        self.Parameterization = np.concatenate((P1,P2,P3))
        


    def theta_to_q( self, theta ):

        phi1, phi2, phi3 = self.get_vhc_funcs()
        P1, P2, P3 = self.get_parameterization()

        q1, q2, q3 = phi1(theta, P1), phi2(theta,P2), phi3(theta,P3)

        Q_DQ = np.array([ [q1],[q2],[q3] ])

        return Q_DQ 

    
    
    
    def get_parameter_str_symbols_dicts( self ):

        """
        returns a ordered dictionary of pvhc-paramaters in the format:
        {string: sympy.symbol}
        the dictionary is ordered alphabetaically with respect to the keys(strings)
        """

        return self.P_1_str_P_1, self.P_2_str_P_2, self.P_3_str_P_3




    def get_parameter_symbols( self ):
        
        """
        return the vhc-functions paramaters as sympy.symbols
        """

        dict_1, dict_2, dict_3 = self.get_parameter_str_symbols_dicts()

        P_1 = dict_1.values()
        P_2 = dict_2.values()
        P_3 = dict_3.values()

        return P_1, P_2, P_3




    def get_parameter_strings( self ):
        
        """
        return the vhc-functions paramaters as strings
        """

        dict_1, dict_2, dict_3 = self.get_parameter_str_symbols_dicts()

        P_1 = dict_1.keys()
        P_2 = dict_2.keys()
        P_3 = dict_3.keys()

        return P_1, P_2, P_3




    def get_sym_vhcs( self ):

        """
        returns symbolic expressions the vhc-functions
        """

        return self.phi_1, self.phi_2, self.phi_3



    def get_sym_vhc_dots( self ):

        """
        returns symbolic expressions for time-differentiated
        phi-expressions
        """

        return self.phi_1_d, self.phi_2_d, self.phi_3_d


    def get_sym_vhc_ddots( self ):

        """
        returns symbolic expressions for twice time-differentiated
        phi-expressions
        """

        return self.phi_1_dd, self.phi_2_dd, self.phi_3_dd



    def get_parameter_dicts( self ):

        """
        returns a dictionaries for vhc-functions in format {string: float}
        containing the parameters and heir current numerical value for writing to file
        """

        self.create_parameter_dicts()

        return self.P_1_P1, self.P_2_P2, self.P_3_P3




    def get_func_vhc( self ):

        """
        returns the lambdified vhc-expressions as callable functions
        phi(theta, P_i)
        """

        return self.phi1, self.phi2, self.phi3



    def get_func_Dphi( self ):

        """
        returns the lambdified vhc-expressions as callable functions
        phi(theta, P_i)
        """

        return self.Dphi1, self.Dphi2, self.Dphi3

    def get_sym_Dphi( self ):

        """
        returns the lambdified vhc-expressions as callable functions
        phi(theta, P_i)
        """

        return self.Dphi_1, self.Dphi_2, self.Dphi_3

    def get_full_parameter_list( self ):

        """
        returns all vhc-parameters as sympy.symbols in list [P_1 P_2 P_3] 
        """
        
        P_1 = self.P_1
        P_2 = self.P_2
        P_3 = self.P_3

        P = []
        P.extend(P_1)
        P.extend(P_2)
        P.extend(P_3)

        return P



    def get_full_parameterization_list( self ):

        """
        returns all numerical values of vhc-parameters in list [P1, P2, P3]
        """

        return self.Parameterization

    def get_full_parameter_dict( self ):

	"""
	returns a dictionary of {sympy.symbol:float} for use in substituting
	"""
	
	parameters = self.get_full_parameter_list()
	parameterization = self.get_full_parameterization_list()


	subs_dict = dict(zip(parameters,parameterization)) 

	return subs_dict




    def get_parameterization( self ):

        """
        returns the vhc-functions numerical parameters seperately
        """

        return self.P1, self.P2, self.P3 



    def get_number_of_parameters( self ):
        
        return len(self.P_1), len(self.P_2), len(self.P_3)



    def parameterization( self, P1, P2, P3 ):

        """
        set the phi-functions parameters numerically
        """
        
        self.P1, self.P2, self.P3 = P1, P2, P3
        
        P = np.concatenate((P1,P2,P3))
        self.Parameterization = P
        


    def alt_parameterization( self, P ):

        """
        set the phi-functions parameters numerically
        """
        
        len1, len2, len3 = self.get_number_of_parameters()
        
        P1, P2, P3 = P[0:len1], P[len1:(len1+len2)], P[(len1+len2):(len1+len2+len3)]
        
        self.P1, self.P2, self.P3 = np.array(P1), np.array(P2), np.array(P3)
        self.Parameterization = P
   
    
    
    
    
    
    





class VHC_Generator( object ):

    """
    creates vhc-structures 
    """

    def __init__( self ): pass
        
    @staticmethod
    def polynomial( equation_number, order ):

        t = symbols('t')
        theta = Function('theta')(t)

        polynomial = 0
        parameters = list()

        for i in range(0, order):
            
            parameter = 'p_' + (str(equation_number) + str(i))
            polynomial += symbols(parameter)*theta**i 
            parameters.append(parameter) 
        
        return polynomial


    @staticmethod
    def trig_polynomial( equation_number, order ):

        t = symbols('t')
        theta = Function('theta')(t)

        polynomial = 0
        parameters = []

        for i in range(1,order):
            
            a_i = symbols('a_' + str(equation_number) + str(i))
            b_i = symbols('b_' + str(equation_number) + str(i))

            aw_i = symbols('aw_' + str(equation_number) + str(i))
            bw_i = symbols('bw_' + str(equation_number) + str(i))

            polynomial += a_i*cos(aw_i*theta) + b_i*sin(bw_i*theta)
            parameters.extend([a_i, aw_i, b_i, bw_i])
       
        return polynomial




